// Jade MIF backend
// Copyright (c) 1997 Kathleen Marszalek
// Version 1.0 (beta 2b)
// Date: October 26, 1997

#ifndef MifDoc_INCLUDED
#define MifDoc_INCLUDED 1

#include "config.h"
#include "IList.h"
#include "Vector.h"
#include "Owner.h"
#include "PointerTable.h"
#include "OutputByteStream.h"
#include "TmpOutputByteStream.h"
#include "StringC.h"
#include "CmdLineApp.h"
#include "Message.h"
#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
//#define assert(exp) (void)( (exp) || (throw "assertion_failed", 0) )

#ifdef DSSSL_NAMESPACE
namespace DSSSL_NAMESPACE {
#endif

#ifdef SP_NAMESPACE
using namespace SP_NAMESPACE;
#endif

#define PROPERTY( propertyType, propertyName ) \
    propertyType propertyName; \
    void set ## propertyName ## ( propertyType p ) \
     { propertyName = p; setProperties |= f ## propertyName ## ;} 

#define PROPERTY_OUT( propertyName ) \
    if( properties & f ## propertyName ) \
        os << '\n' << MifOutputByteStream::INDENT \
           << "<" #propertyName " " << propertyName << ">"; 

#define VECTOR_OF_PROPERTIES( propertyType, propertyName ) \
    Vector< ## propertyType ## > propertyName ## s; \
    void propertyName ## sAreSet() \
     { setProperties |= f ## propertyName ## s ##;}

#define STATIC_PROPERTY( propertyType, propertyName ) \
    static propertyType propertyName; \
    static void set ## propertyName ## ( propertyType p ) \
     { propertyName = p; setProperties |= f ## propertyName ## ;}

class StringHash {
public:
  static unsigned long hash( const String<char> & );
};

class MifOutputByteStream;
class MifTmpOutputByteStream;
class MifDoc {

  public:

    MifDoc( const String<CmdLineApp::AppChar> &fileLoc, CmdLineApp *app );
    ~MifDoc();
    static MifDoc &curInstance() { assert( CurInstance != NULL ); return *CurInstance; }
    static MifDoc *CurInstance;

    struct T_indent {

        T_indent( unsigned i ) : data( i ) {}
        long data;
        operator unsigned() const { return data; }
    };

    struct T_dimension {

        T_dimension( long u = 0 ) : data( u ) {}
        long data;
        operator long() const { return data; }
        T_dimension &operator=( long s ) { data = s; return *this; }
    };

    struct T_string : public String<char> {

        T_string() : String<char>() {};
        T_string( const String<char> &s ) : String<char>( s ) {};
        T_string( String<CmdLineApp::AppChar> &s ) : String<char>() 
          { for( size_t i = 0; i < s.size(); i++ ) *this += char( s[i] ); }
        T_string( const char *cS ) : String<char>( cS, strlen( cS ) ) {};
        T_string &operator=( String<char> s ) { String<char>::operator=( s ); return *this; }
    };

    struct T_boolean {

        T_boolean( bool b = false ) : data( b ) {}
        bool data;
        operator bool() const { return data; }
        T_boolean &operator=( bool b ) { data = b; return *this; }
    };

    typedef T_string        T_tagstring;
    typedef T_string        T_pathname;
    typedef String<char>    T_keyword;

    typedef long        T_integer;
    typedef unsigned    T_ID;
    typedef long        T_degrees;
    typedef long        T_percentage;
    typedef long        T_metric;

    struct T_WH { T_dimension w, h; // width, height      
                  T_WH( T_dimension w_, T_dimension h_ ) : w( w_ ), h( h_ ) {}  
                  T_WH() : w( 0L ), h( 0L ) {} };
    struct T_XY { T_dimension x, y;
                  T_XY( T_dimension x_, T_dimension y_ ) : x( x_ ), y( y_ ) {}  
                  T_XY() : x( 0L ), y( 0L ) {} };
    struct T_LTRB { T_dimension l, t, r, b; // left, top, right, bottom
                    T_LTRB( T_dimension l_, T_dimension t_, T_dimension r_, T_dimension b_ )
                     : l( l_ ), t( t_ ), r( r_ ), b( b_ ) {}
                    T_LTRB() : l( 0L ), t( 0L ), r( 0L ), b( 0L ) {} };
    struct T_LTWH { T_dimension l, t, w, h; // left, top, width, height
                    T_LTWH( T_dimension l_, T_dimension t_, T_dimension w_, T_dimension h_ )
                     : l( l_ ), t( t_ ), w( w_ ), h( h_ ) {}
                    T_LTWH() : l( 0L ), t( 0L ), w( 0L ), h( 0L ) {} };
    struct T_XYWH { T_dimension x, y, w, h; };          
 

    // special characters
    static const String<char> sTab;
    static const String<char> sHardSpace;
    static const String<char> sSoftHyphen;
    static const String<char> sHardHyphen;
    static const String<char> sHardReturn;
    static const String<char> sNumberSpace;
    static const String<char> sThinSpace;
    static const String<char> sEnSpace;
    static const String<char> sEmSpace;
    static const String<char> sCent;
    static const String<char> sPound;
    static const String<char> sYen;
    static const String<char> sEnDash;
    static const String<char> sEmDash;
    static const String<char> sDagger;
    static const String<char> sDoubleDagger;
    static const String<char> sBullet;

    // for tagged text flows
    static const String<char> sA;

    // FWeight
    static const String<char> sRegular;
    static const String<char> sBold;

    // FAngle
    // sRegular;
    static const String<char> sItalic;

    // FVar
    // sRegular;

    // FColor
    static const String<char> sWhite;
    static const String<char> sBlack;

    // FUnderlining
    static const String<char> sFNoUnderlining;
    static const String<char> sFSingle;
    static const String<char> sFDouble;
    static const String<char> sFNumeric;

    // FPosition
    static const String<char> sFNormal;
    static const String<char> sFSuperscript;
    static const String<char> sFSubscript;

    // FCase
    static const String<char> sFAsTyped;
    static const String<char> sFSmallCaps;
    static const String<char> sFLowercase;
    static const String<char> sFUppercase;

    // FFamily
    static const String<char> sTimesNewRoman;

    struct FontFormat : public Link {

        enum { fNone = 0, fFFamily = 0x1, fFAngle = 0x2, fFWeight = 0x4, fFVar = 0x8,
               fFSize = 0x10, fFColor = 0x20, fFUnderlining = 0x40, fFOverline = 0x80,
               fFStrike = 0x100, fFPosition = 0x200, fFPairKern = 0x400, fFCase = 0x800,
               fFDX = 0x1000, fFDY = 0x2000, fFDW = 0x4000, fAll = 0x7FFF };
        unsigned setProperties;

        FontFormat() : setProperties( 0 ) {}

        // font name
        PROPERTY( T_string, FFamily );
        PROPERTY( T_string, FAngle );
        PROPERTY( T_string, FWeight );
        PROPERTY( T_string, FVar );

        // font size and color
        PROPERTY( T_dimension, FSize );
        PROPERTY( T_tagstring, FColor );
                
        // font style
        PROPERTY( T_keyword, FUnderlining );
        PROPERTY( T_boolean, FOverline );
        PROPERTY( T_boolean, FStrike );
        PROPERTY( T_keyword, FPosition );
        PROPERTY( T_boolean, FPairKern );
        PROPERTY( T_keyword, FCase );

        // kerning information
        PROPERTY( T_percentage, FDX );
        PROPERTY( T_percentage, FDY );
        PROPERTY( T_percentage, FDW );

        void setDSSSLDefaults() {
        
            setFFamily( sTimesNewRoman );
            setFAngle( sRegular );
            setFWeight( sRegular );
            setFVar( sRegular );
            setFSize( 10000 );
            // setFColor( sNONE ),
            setFUnderlining( sFNoUnderlining );
            setFOverline( false );
            setFStrike( false );
            setFPosition( sFNormal );
            setFPairKern( false );
            setFCase( sFAsTyped );
            setFDX( 0 );
            setFDY( 0 );
            setFDW( 0 );
        }

        enum FontStatement { stPgfFont, stFont };

        unsigned compare( FontFormat &f );
        void setFrom( FontFormat &f, unsigned properties );
        void out( MifOutputByteStream &os, unsigned properties, FontStatement fontStatement );
        void ffOut( MifOutputByteStream &os, unsigned properties, FontStatement fontStatement )
              { out( os, properties, fontStatement ); }
        void out( MifOutputByteStream &os, FontStatement fontStatement )
            { out( os, setProperties, fontStatement ); }
        void ffOut( MifOutputByteStream &os, FontStatement fontStatement )
            { out( os, fontStatement ); }
        void updateFrom( FontFormat & );
        void ffUpdateFrom( FontFormat &f ) { updateFrom( f ); }
        void clearSetProperties() { setProperties = 0; }
    };

    static const String<char> sNONE;

    // paragraph formats
    static const String<char> sDefaultPgfFormat;
    static const String<char> sHeader;
    static const String<char> sFooter;

    // PgfAlignment
    static const String<char> sLeftRight;
    static const String<char> sLeft;
    static const String<char> sRight;
    static const String<char> sCenter;

    // PgfLineSpacing
    static const String<char> sFixed;
    static const String<char> sProportional;

    // PgfCellAlignment
    static const String<char> sTop;
    static const String<char> sMiddle;
    static const String<char> sBottom;

    // PgfPlacement
    static const String<char> sAnywhere;
    static const String<char> sColumnTop;
    static const String<char> sPageTop;
    static const String<char> sLPageTop;
    static const String<char> sRPageTop;

    // PgfPlacementStyle
    static const String<char> sNormal;
    static const String<char> sStraddle;

    // PgfLanguage
    static const String<char> sNoLanguage;
    static const String<char> sUSEnglish;
    static const String<char> sUKEnglish;
    static const String<char> sGerman;
    static const String<char> sSwissGerman;
    static const String<char> sFrench;
    static const String<char> sCanadianFrench;
    static const String<char> sSpanish;
    static const String<char> sCatalan;
    static const String<char> sItalian;
    static const String<char> sPortuguese;
    static const String<char> sBrazilian;
    static const String<char> sDanish;
    static const String<char> sDutch;
    static const String<char> sNorwegian;
    static const String<char> sNynorsk;
    static const String<char> sFinnish;
    static const String<char> sSwedish;


    // TStype
    // sLeft;
    // sRight;
    // sCenter;
    static const String<char> sDecimal;

    // TSLeaderStr;
    static const String<char> sSPACE;

    struct TabStop {

        enum { fNone = 0, fTSX = 0x1, fTSType = 0x2, fTSLeaderStr = 0x4,
               fTSDecimalChar = 0x8, fAll = 0xF };
        unsigned setProperties;

        TabStop( T_keyword type = sLeft, T_dimension x = 0, T_string tSLeaderStr = sSPACE )
         : TSX( x ), TSType( type ), setProperties( fTSType | fTSX )
          { setTSLeaderStr( tSLeaderStr ); }

        PROPERTY( T_dimension, TSX );
        PROPERTY( T_keyword, TSType );
        PROPERTY( T_string, TSLeaderStr );
        PROPERTY( T_integer, TSDecimalChar );

        void out( MifOutputByteStream &os );
    };

    struct ParagraphFormat : public FontFormat {
    
		enum { fNone = 0, fPgfFIndent = 0x1, fPgfLIndent = 0x2, fPgfRIndent = 0x4,
               fPgfAlignment = 0x8, fPgfSpBefore = 0x10, fPgfSpAfter = 0x20,
               fPgfLineSpacing = 0x40, fPgfLeading = 0x80, fPgfWithPrev = 0x100,
               fPgfWithNext = 0x200, fPgfBlockSize = 0x400, fPgfAutoNum = 0x800,
               fPgfNumFormat = 0x1000, fPgfNumberFont = 0x2000, fPgfHyphenate = 0x4000,
               fHyphenMaxLines = 0x8000, fHyphenMinPrefix = 0x10000,
               fHyphenMinSuffix = 0x20000, fHyphenMinWord = 0x40000,
               fPgfLetterSpace = 0x80000, fPgfLanguage = 0x100000,
               fPgfCellAlignment = 0x200000, fPgfCellMargins = 0x400000,
               fPgfCellLMarginFixed = 0x800000, fPgfCellTMarginFixed = 0x1000000,
               fPgfCellRMarginFixed = 0x2000000,  fPgfCellBMarginFixed = 0x4000000,
               fPgfTag = 0x8000000, fTabStops = 0x10000000, fPgfPlacement = 0x20000000,
               fPgfNumTabs = 0x40000000, fPgfPlacementStyle = 0x80000000,
               fAll = 0xFFFFFFFF };
        unsigned setProperties;

        ParagraphFormat() : setProperties( 0 ) {}

        // basic properties
        PROPERTY( T_tagstring, PgfTag );
        PROPERTY( T_dimension, PgfFIndent );
        PROPERTY( T_dimension, PgfLIndent );
        PROPERTY( T_dimension, PgfRIndent );
        PROPERTY( T_keyword, PgfAlignment );
        PROPERTY( T_dimension, PgfSpBefore );
        PROPERTY( T_dimension, PgfSpAfter );
        PROPERTY( T_keyword, PgfLineSpacing );
        PROPERTY( T_dimension, PgfLeading );
        PROPERTY( T_integer, PgfNumTabs );
        Vector<TabStop> TabStops;

        // default font properties

        // pagination properties
        PROPERTY( T_keyword, PgfPlacement );
        PROPERTY( T_keyword, PgfPlacementStyle );
        PROPERTY( T_boolean, PgfWithPrev );
        PROPERTY( T_boolean, PgfWithNext );
        PROPERTY( T_integer, PgfBlockSize );

        // numbering properties
        PROPERTY( T_boolean, PgfAutoNum );
        PROPERTY( T_string, PgfNumFormat );
        PROPERTY( T_tagstring, PgfNumberFont );

        // advanced properties
        PROPERTY( T_boolean, PgfHyphenate );
        PROPERTY( T_integer, HyphenMaxLines );
        PROPERTY( T_integer, HyphenMinPrefix );
        PROPERTY( T_integer, HyphenMinSuffix );
        PROPERTY( T_integer, HyphenMinWord );
        PROPERTY( T_boolean, PgfLetterSpace );
        PROPERTY( T_keyword, PgfLanguage );
            
        // table cell properties
        PROPERTY( T_keyword, PgfCellAlignment );
        PROPERTY( T_LTRB, PgfCellMargins );
        PROPERTY( T_boolean, PgfCellLMarginFixed );
        PROPERTY( T_boolean, PgfCellTMarginFixed );
        PROPERTY( T_boolean, PgfCellRMarginFixed );
        PROPERTY( T_boolean, PgfCellBMarginFixed );

        // miscellaneous properties
        // none

        void setDSSSLDefaults() {
        
            FontFormat::setDSSSLDefaults();
            setPgfFIndent( 0 );
            setPgfLIndent( 0 );
            setPgfRIndent( 0 );
            setPgfAlignment( sLeft );
            setPgfSpBefore( 0 );
            setPgfSpAfter( 0 );
            setPgfLineSpacing( sFixed );
            setPgfLeading( 0 );
            setPgfNumTabs( 0 );
            setPgfWithPrev( false );
            setPgfWithNext( false );
            setPgfBlockSize( 2 );
            setPgfAutoNum( false );
            // setPgfNumFormat( sNONE );
            // setPgfNumberFont( sNONE );
            setPgfHyphenate( false );
            setHyphenMaxLines( 999 );
            setHyphenMinPrefix( 2 );
            setHyphenMinSuffix( 2 );
            setHyphenMinWord( 2 );
            setPgfLetterSpace( false );
            setPgfLanguage( sNoLanguage );
            setPgfCellAlignment( sTop );
            setPgfCellMargins( T_LTRB( 0, 0, 0, 0 ) );
            setPgfCellLMarginFixed( true );
            setPgfCellTMarginFixed( true );
            setPgfCellRMarginFixed( true );
            setPgfCellBMarginFixed( true );
            // setPgfTag( sNONE );
            setPgfPlacement( sAnywhere );
            setPgfPlacementStyle( sNormal );
        }

        void forceSetProperties( unsigned properties, unsigned fontProperties )
            { setProperties = properties; FontFormat::setProperties = fontProperties; }
        unsigned &ffSetProperties() { return FontFormat::setProperties; }
        unsigned compare( ParagraphFormat & );
        void setFrom( ParagraphFormat &, unsigned, unsigned );
        void out( MifOutputByteStream &os, unsigned, unsigned,
                  bool excludeCellProperties = true );
        void out( MifOutputByteStream &os, bool excludeCellProperties = true )
            { out( os, setProperties, (unsigned)FontFormat::setProperties,
                   excludeCellProperties ); }
        void updateFrom( ParagraphFormat & );
        void clearSetProperties() { setProperties = 0; FontFormat::clearSetProperties(); }
    };

    // DPageNumStyle
    static const String<char> sArabic;
    static const String<char> sUCRoman;
    static const String<char> sLCRoman;
    static const String<char> sUCAlpha;
    static const String<char> sLCAlpha;

    // DParity
    static const String<char> sFirstLeft;
    static const String<char> sFirstRight;

    struct Document {

        enum { fNone = 0, fDPageSize = 0x1, fDStartPage = 0x2, fDPageNumStyle = 0x4,
               fDTwoSides = 0x8, fDParity = 0x10, fDMargins = 0x20, fDColumns = 0x40,
               fDColumnGap = 0x80, fAll = 0xFF };
        unsigned setProperties;

        Document() : setProperties( fNone ) {}

        // column properties
        PROPERTY( T_LTRB, DMargins );
        PROPERTY( T_integer, DColumns );
        PROPERTY( T_dimension, DColumnGap );
        PROPERTY( T_WH, DPageSize );

        // pagination
        PROPERTY( T_integer, DStartPage );
        PROPERTY( T_keyword, DPageNumStyle );
        PROPERTY( T_boolean, DTwoSides );
        PROPERTY( T_keyword, DParity );

        void out( MifOutputByteStream &os );
    };

    struct Object {

        enum { fNone = 0, fID = 0x1, fPen = 0x2, fFill = 0x4, fPenWidth = 0x8,
               fObjectNext = 0x10 };
        unsigned setProperties;

        Object( T_integer pen = 15, T_integer fill = 15, T_dimension penWidth = 0 )
         : setProperties( 0 )
         { setID( ++IDCnt ); setPen( pen ); setFill( fill ); setPenWidth( penWidth ); }

        static T_ID IDCnt;
        PROPERTY( T_integer, ID );
        PROPERTY( T_integer, Pen );
        PROPERTY( T_integer, Fill );
        PROPERTY( T_dimension, PenWidth );

        void outObjectProperties( MifOutputByteStream &os );
        virtual void out( MifOutputByteStream &os ) = 0;
    };

    // HeadCap/TailCap
    static const String<char> sButt;
    static const String<char> sRound;
    static const String<char> sSquare;
    
    struct PolyLine : public Object {

        enum { fHeadCap = fObjectNext << 1, fTailCap = fHeadCap << 1 };

        PolyLine( T_keyword cap, T_integer pen = 15, T_integer fill = 15,
                  T_dimension penWidth = 0 )
         : Object( pen, fill, penWidth ) { setHeadCap( cap ); setTailCap( cap ); }

        PROPERTY( T_keyword, HeadCap );
        PROPERTY( T_keyword, TailCap );
        Vector<T_XY> Points;

        void out( MifOutputByteStream &os );
    };
    
    struct ImportObject : public Object {

        enum { fImportObFileDI = fObjectNext << 1, fBitMapDpi = fImportObFileDI << 1,
               fShapeRect = fBitMapDpi << 1, fImportObFixedSize = fShapeRect << 1,
               fNativeOrigin = fImportObFixedSize << 1 };

        ImportObject( T_pathname importObFileDI, T_LTWH shapeRect,
                      T_boolean importObFixedSize = true, T_integer bitMapDpi = 72 ) : Object()
         { setImportObFileDI( importObFileDI ); setShapeRect( shapeRect );
           setImportObFixedSize( importObFixedSize ); setBitMapDpi( bitMapDpi ); }

        PROPERTY( T_pathname, ImportObFileDI );
        PROPERTY( T_integer, BitMapDpi );
        PROPERTY( T_LTWH, ShapeRect );
        PROPERTY( T_boolean, ImportObFixedSize );
        PROPERTY( T_XY, NativeOrigin );

        void out( MifOutputByteStream &os );
    };

    // FrameType
    static const String<char> sInline;
    static const String<char> sRunIntoParagraph;
    static const String<char> sBelow;

    // AnchorAlign
    // Left
    // Center
    // Right
    // Inside
    // Outside

    struct Frame : public Object {

        enum { fShapeRect = fObjectNext << 1, fFrameType = fShapeRect << 1, 
               fNSOffset = fFrameType << 1, fBLOffset = fNSOffset << 1, 
               fAnchorAlign = fBLOffset << 1 };

        Frame() : Object() {}

        PROPERTY( T_LTWH, ShapeRect );
        PROPERTY( T_keyword, FrameType );
        PROPERTY( T_dimension, NSOffset );
        PROPERTY( T_dimension, BLOffset );
        PROPERTY( T_keyword, AnchorAlign );
        Vector<Object*> Objects;

        void out( MifOutputByteStream &os );
    };

    struct TextRect : public Object {

        enum { fID = fObjectNext << 1, fShapeRect = fID << 1, fTRNumColumns = fShapeRect << 1,
               fTRColumnGap = fTRNumColumns << 1, fTRColumnBalance = fTRColumnGap << 1  };

        TextRect() : Object()  {}
        TextRect( T_LTWH shapeRect, T_integer tRNumColumns = 1, T_dimension tRColumnGap = 0,
                  bool tRColumnBalance = false )
         : Object() { setShapeRect( shapeRect ); setTRNumColumns( tRNumColumns );
                      setTRColumnGap( tRColumnGap ); setTRColumnBalance( tRColumnBalance ); }

        PROPERTY( T_LTWH, ShapeRect );
        PROPERTY( T_integer, TRNumColumns );
        PROPERTY( T_dimension, TRColumnGap );
        PROPERTY( T_boolean, TRColumnBalance );

        void out( MifOutputByteStream &os );
    };

    // PageType
    static const String<char> sLeftMasterPage;
    static const String<char> sRightMasterPage;
    static const String<char> sOtherMasterPage;
    static const String<char> sBodyPage;

    // PageTag
    static const String<char> sFirst;
    // sLeft
    // sRight

    struct Page {

        enum { fNone = 0, fPageType = 0x1, fPageTag = 0x2, fPageBackground = 0x4 };
        unsigned setProperties;

        Page( T_keyword pageType, T_tagstring pageTag = sNONE,
              T_tagstring pageBackground = sNONE )
         : PageType( pageType ), PageTag( pageTag ), PageBackground( pageBackground ),
           setProperties( fPageType | fPageTag )
         { if( PageTag != sNONE ) setProperties |= fPageTag;
           if( PageBackground != sNONE ) setProperties |= fPageBackground; }
        Page() : PageType( MifDoc::sRightMasterPage ), setProperties( fPageType ) {}

        PROPERTY( T_keyword, PageType );
        PROPERTY( T_tagstring, PageTag );
        PROPERTY( T_tagstring, PageBackground ); // keyword in MIF
        Vector<TextRect> TextRects;

        void out( MifOutputByteStream &os );
    };

    struct TagStream {

        enum TagStreamClass { tsTagStream, tsTextFlow, tsCell, tsPara } TagStreamClass;

        TagStream( T_indent osIndent = 2 );
        ~TagStream();

        T_tagstring InitialPgfTag;
        bool PgfTagUsed;

        ParagraphFormat &format() { return Format; }
        MifTmpOutputByteStream &content() { assert( Content != NULL ); return *Content; }
        void setParagraphFormat( ParagraphFormat &pf ) { Format = pf; }

      protected:
        MifTmpOutputByteStream *Content;               
        ParagraphFormat Format;
    };

    struct TextFlow : public TagStream {

        enum { fNone = 0, fTFTag = 0x1, fTFAutoConnect = 0x02 };
        unsigned setProperties;

        TextFlow() : setProperties( fNone ), TextRectIDUsed( false )
         { TagStreamClass = tsTextFlow; }
        TextFlow( const TextRect &textRect, bool body,
                  ParagraphFormat *format = NULL, T_tagstring pgfTag = sNONE );

        T_ID TextRectID; // for ParaLine (not in MIF)
        bool TextRectIDUsed;
        PROPERTY( T_tagstring, TFTag );
        PROPERTY( T_boolean, TFAutoConnect );

        const TextRect *textRect() { assert( TextRect_ != NULL ); return TextRect_; }
        void out( MifOutputByteStream &os, bool resolveCrossReferences = false );

      protected:
        const TextRect *TextRect_;
    };

    struct Para : public TagStream {

        enum { fNone = 0, fPgfTag = 0x1, fParagraphFormat = 0x2 };
        unsigned setProperties;

        Para( T_indent osIndent = 2 )
         : TagStream( osIndent ), setProperties( fParagraphFormat ), leaderTabsSet( false )
          { TagStreamClass = tsPara; }

        static bool currentlyOpened;
        bool leaderTabsSet;
        PROPERTY( T_tagstring, PgfTag );

        ParagraphFormat &curFormat() { return CurFormat; }

        void out( MifOutputByteStream &os );
        void outProlog( MifOutputByteStream &os );

        static void outSimpleProlog( MifOutputByteStream &os );
        static void outEpilog( MifOutputByteStream &os );

      protected:
        ParagraphFormat CurFormat;

    };

    struct ParaLine {

        enum { fNone = 0, fTextRectID = 0x1, fATbl = 0x2 };
        static unsigned setProperties;

        STATIC_PROPERTY( T_ID, TextRectID );
        STATIC_PROPERTY( T_ID, ATbl );

        static void outProlog( MifOutputByteStream &os );
        static void outEpilog( MifOutputByteStream &os );
    };

    struct PgfCatalog {

        Vector<ParagraphFormat> ParaFormats;
        void out( MifOutputByteStream &os );
    };

    // TblAlignment
    // Left
    // Center
    // Right
    static const String<char> sInside;
    static const String<char> sOutside;

    struct Ruling {

        enum { fNone = 0, fRulingTag = 0x1, fRulingPenWidth = 0x2, fRulingGap = 0x4,
               fRulingPen = 0x8, fRulingLines = 0x10 };
        unsigned setProperties;

        Ruling() : setProperties( 0 ), RulingPenWidth( 1000 ), RulingGap( 1000 ),
                   RulingPen( 0 ), RulingLines( 1 ) {};
        Ruling( T_tagstring RulingTag_ ) 
         : setProperties( 0 ), RulingPenWidth( 1000 ), RulingGap( 1000 ),
           RulingPen( 0 ), RulingLines( 1 )
        { setRulingTag( RulingTag_ ); }

        PROPERTY( T_tagstring, RulingTag );
        PROPERTY( T_dimension, RulingPenWidth );
        PROPERTY( T_dimension, RulingGap );
        PROPERTY( T_integer, RulingPen );
        PROPERTY( T_integer, RulingLines );

        static const String<char> &key( Ruling & );
        void out( MifOutputByteStream &os ) const;
        String<char> Key;        
      
    };

    struct RulingCatalog {

        PointerTable<Ruling*,String<char>,StringHash,Ruling> Rulings;
        void out( MifOutputByteStream &os );
    };

    struct Cell : public TagStream {

        enum { fNone = 0, fCellLRuling = 0x1, fCellBRuling = 0x2, fCellRRuling = 0x4,
               fCellTRuling = 0x8, fCellColumns = 0x10, fCellRows = 0x20 };
        unsigned setProperties;

        Cell( T_indent osIndent = 12 )
         : TagStream( osIndent ), setProperties( 0 ) { TagStreamClass = tsCell; }

        PROPERTY( T_tagstring, CellLRuling );
        PROPERTY( T_tagstring, CellBRuling );
        PROPERTY( T_tagstring, CellRRuling );
        PROPERTY( T_tagstring, CellTRuling );
        PROPERTY( T_integer, CellColumns );
        PROPERTY( T_integer, CellRows );

        void out( MifOutputByteStream &os, bool resolveCrossReferences = false );
    };

    struct Row {

        Vector<Cell> Cells;

        void out( MifOutputByteStream &os, bool resolveCrossReferences = false );
    };

    struct TblColumn {

        enum { fNone = 0, fTblColumnNum = 0x1, fTblColumnWidth = 0x2 };
        unsigned setProperties;

        TblColumn() : setProperties( 0 ), TblColumnNum( 0 ), TblColumnWidth( 0 ) {};
        TblColumn( T_integer TblColumnNum_, T_dimension TblColumnWidth_ )
         : setProperties( 0 )
         { setTblColumnNum( TblColumnNum_ ); setTblColumnWidth( TblColumnWidth_ ); }

        PROPERTY( T_integer, TblColumnNum );
        PROPERTY( T_dimension, TblColumnWidth );

        void out( MifOutputByteStream &os );
    };

    // TblTitlePlacement
    static const String<char> sNone;

    struct TblFormat {

        enum { fNone = 0, fTblTag = 0x1, fTblLIndent = 0x2, fTblRIndent = 0x4,
               fTblSpBefore = 0x8, fTblSpAfter = 0x10, fTblAlignment = 0x20,
               fTblCellMargins = 0x40, fTblWidth = 0x80, fTblTitlePlacement = 0x100,
               fAll = 0x1ff };
        unsigned setProperties;

        TblFormat() : setProperties( 0 ) {}
        TblFormat( T_tagstring TblTag_ ) : setProperties( 0 ) { setTblTag( TblTag_ ); }

        void setDSSSLDefaults() {
        
            setTblLIndent( 0 );
            setTblRIndent( 0 );
            setTblSpBefore( 0 );
            setTblSpAfter( 0 );
            setTblAlignment( sLeft );
            setTblCellMargins( T_LTRB( 0, 0, 0, 0 ) );
            setTblTitlePlacement( sNone );
        }

        // basic properties
        PROPERTY( T_tagstring, TblTag );
        PROPERTY( T_dimension, TblLIndent );
        PROPERTY( T_dimension, TblRIndent );
        PROPERTY( T_dimension, TblSpBefore );
        PROPERTY( T_dimension, TblSpAfter );
        PROPERTY( T_keyword, TblAlignment );
        PROPERTY( T_LTRB, TblCellMargins );
        PROPERTY( T_keyword, TblTitlePlacement );

        // column properties
        PROPERTY( T_dimension, TblWidth );
        Vector<TblColumn> TblColumns;

        void out( MifOutputByteStream &os );
    };

    struct TblCatalog {

        Vector<TblFormat> TblFormats;
        void out( MifOutputByteStream &os );
    };

    // TblTag
    static const String<char> sDefaultTblFormat;

    struct Tbl {

        enum { fNone = 0, fTblID = 0x1, fTblTag = 0x2, fTblNumColumns = 0x4,
               fTblColumnWidths = 0x8, fTblFormat = 0x10 };
        unsigned setProperties;

        Tbl() : setProperties( 0 ), TblIDUsed( false )
          { setTblID( MifDoc::CurInstance->nextID() ); setTblTag( MifDoc::sDefaultTblFormat ); }

        PROPERTY( T_ID, TblID );
        PROPERTY( T_tagstring, TblTag );
        bool TblIDUsed;
        TblFormat TblFormat;

        // table columns
        PROPERTY( T_integer, TblNumColumns );
        VECTOR_OF_PROPERTIES( T_dimension, TblColumnWidth );

        // table rows
        Vector<Row> TblH;
        Vector<Row> TblBody;
        Vector<Row> TblF;

        void out( MifOutputByteStream &os, bool resolveCrossReferences = false );
    };

    // paragraph formats
    static const String<char> sPageNumXRefFormatName;

    struct XRefFormat {

        enum { fXRefName = 0x1, fXRefDef = 0x2 };
        unsigned setProperties;

        XRefFormat() {}
        XRefFormat( T_string XRefName_, T_string XRefDef_ )
         { setXRefName( XRefName_ ); setXRefDef( XRefDef_ ); }

        PROPERTY( T_string, XRefName );
        PROPERTY( T_string, XRefDef );

        void out( MifOutputByteStream &os );
    };

    struct XRef {

        enum { fXRefName = 0x1, fXRefSrcFile = 0x2, fXRefText = 0x4, fXRefSrcText };
        unsigned setProperties;

        XRef() {}
        XRef( T_tagstring XRefName_, T_string XRefSrcText_, T_string XRefText_,
              T_pathname XRefSrcFile_ )
         { setXRefName( XRefName_ ); setXRefSrcText( XRefSrcText_ ); setXRefText( XRefText_ );
           setXRefSrcFile( XRefSrcFile_ ); }

        PROPERTY( T_tagstring, XRefName );
        PROPERTY( T_string, XRefSrcText );
        PROPERTY( T_pathname, XRefSrcFile );
        PROPERTY( T_string, XRefText );

        void out( MifOutputByteStream &os );
    };

    struct Marker {

        enum { fMType = 0x1, fMText = 0x2 };
        unsigned setProperties;

        Marker( T_string MText_ ) { setMText( MText_ ); setMType( 9 ); }

        PROPERTY( T_integer, MType );
        PROPERTY( T_string, MText );

        static T_string makeText( unsigned long groveIndex, unsigned long elementIndex );
        void out( MifOutputByteStream &os );
    };

    struct BookComponent {
    
        BookComponent() : epilogOs( NULL ), pageNumXRefFormatGenerated( false ) {}
        BookComponent( MifTmpOutputByteStream *epilogOs );
        void commit( const String<CmdLineApp::AppChar> &dirName,
                     bool resolveCrossReferences = true );
        
        String<CmdLineApp::AppChar> FileName;
        Vector<XRefFormat> XRefFormats;
        Document Document;
        PgfCatalog PgfCatalog;
        TblCatalog TblCatalog;
        Vector<Frame> AFrames;
        Vector<Tbl> Tbls;
        Vector<Page> Pages;
        Vector<TextFlow> TextFlows;
        RulingCatalog RulingCatalog;

        bool pageNumXRefFormatGenerated;

        MifTmpOutputByteStream *epilogOs;             
    };

    static char escapeChar() { return '\x0'; }

    struct PotentialMarkerInfo {

        PotentialMarkerInfo() {}
        PotentialMarkerInfo
         ( unsigned long groveIndex_, unsigned long elementIndex_, int tagIndent_ )
          : groveIndex( groveIndex_ ), elementIndex( elementIndex_ ), tagIndent( tagIndent_ ) {}
        
        unsigned long groveIndex;
        unsigned long elementIndex;
        int tagIndent;
        static const char TypeChar;
    };

    struct XRefInfo {

        XRefInfo() {}
        XRefInfo( unsigned long groveIndex_, unsigned long elementIndex_,
                  int tagIndent_ )
          : groveIndex( groveIndex_ ), elementIndex( elementIndex_ ), tagIndent( tagIndent_ ) {}

        unsigned long groveIndex;
        unsigned long elementIndex;
        int tagIndent;
        static const char TypeChar;
    };

    struct ElementSet {

        static const unsigned short REFERENCED_F() { return 0x8000; }
        static const unsigned short BOOK_COMPONENT_INDEX_M() { return 0x7FFF; }
        ElementSet();
        void setReferencedFlag( unsigned long groveIndex, unsigned long n );
        void setBookComponentIndex( unsigned long groveIndex, unsigned long n, int i );
        bool hasBeenReferenced( unsigned long groveIndex, unsigned long n ) const;
        int bookComponentIndex( unsigned long groveIndex, unsigned long n ) const;
      private:
        void add( unsigned long groveIndex, unsigned long n, unsigned short value  );
        bool getValue( unsigned long groveIndex, unsigned long n, unsigned short &result ) const;
        Vector<Vector<unsigned short> > v_;
    };

    void commit();

    void outTagEnd();

    void outHexChar( unsigned code, MifOutputByteStream *o = NULL );
    void outHexChar( unsigned code, String<char> &targetString );
    void outSpecialChar( const String<char> &charName, MifOutputByteStream *o = NULL ); 
    void beginParaLine();
    void endParaLine();
    void outPageNumber();
    void outBreakingPara( T_keyword pgfPlacement );
    void outAFrame( T_ID ID, MifOutputByteStream &os );

    Vector<BookComponent> &bookComponents() { return BookComponents; };
    BookComponent &bookComponent()
     { assert( bookComponents().size() > 0 ); return bookComponents().back(); }

    void setCurOs( MifOutputByteStream &os ) { curOs = &os; } 
    MifOutputByteStream &os() { assert( curOs != NULL ); return *curOs; }

    void setCurTextFlow( TextFlow *tf ) { CurTextFlow = tf; } 
    TextFlow *curTextFlow() { return CurTextFlow; }

    void setCurCell( Cell *c ) { CurCell = c; } 
    Cell *curCell() { return CurCell; }

    void setCurTblNum( size_t n ) { CurTblNum = n; } 
    Tbl *curTbl( bool assertNotNull = true )
     { if( assertNotNull ) assert( CurTblNum > 0 );
       return CurTblNum > 0 ? &bookComponent().Tbls[CurTblNum-1] : NULL; }

    void setCurPara( Para *p ) {  CurPara = p; } 
    Para *curPara( bool assertNotNull = true )
     { if( assertNotNull ) assert( CurPara != NULL ); return CurPara; }

    void setCurParagraphFormat( ParagraphFormat &pf ) { CurFormat = pf; } 
    ParagraphFormat &curFormat() { return CurFormat; }

    TagStream &curTagStream()
      { assert( TagStreamStack.size() > 0 ); return *TagStreamStack.back(); }

    void enterTextFlow( TextFlow &textFlow  ) { enterTagStream( textFlow ); }
    void exitTextFlow() { setCurTextFlow( NULL ); exitTagStream(); }

    void enterTableCell( Cell &cell ) { enterTagStream( cell ); }
    void exitTableCell() { setCurCell( NULL ); exitTagStream(); }

    void enterPara( Para &p ) { enterTagStream( p ); }
    void exitPara() { setCurPara( NULL ); exitTagStream(); }

    void enterBookComponent();
    void exitBookComponent();
    void enterTagStream( TagStream &tagStream );
    void exitTagStream();
    T_ID nextID() { return ++NextID; }

    Document &document() { return bookComponent().Document; }
    PgfCatalog &pgfCatalog() { return bookComponent().PgfCatalog; }
    TblCatalog &tblCatalog() { return bookComponent().TblCatalog; }
    RulingCatalog &rulingCatalog() { return bookComponent().RulingCatalog; }
    Vector<Tbl> &tbls() { return bookComponent().Tbls; }
    Vector<Page> &pages() { return bookComponent().Pages; }
    Vector<TextFlow> &textFlows() { return bookComponent().TextFlows; }
    Vector<Frame> &aFrames() { return bookComponent().AFrames; }
    ElementSet &elements() { return Elements; }

    const String<CmdLineApp::AppChar> &rootOutputFileLoc() { return RootOutputFileLoc; }

    CmdLineApp *App;

  protected:

    ElementSet Elements;
    void switchToTagStream( TagStream &tagStream, bool startWithDefaultPgfFormat = true );

    String<CmdLineApp::AppChar> RootOutputFileLoc;
    T_ID NextID;

    MifOutputByteStream *curOs;    

    Vector<BookComponent> BookComponents;

    size_t CurTblNum;
    ParagraphFormat CurFormat;
    TextFlow *CurTextFlow;
    Cell *CurCell;
    Para *CurPara;

    Vector<TagStream*> TagStreamStack;
};

class MifOutputByteStream {

    OutputByteStream *os;

  public:
    
    int CurTagIndent;
    static MifDoc::T_indent INDENT;

    void indent() { CurTagIndent += 2; }
    void undent() { CurTagIndent -= 2; assert( CurTagIndent >= 0 ); }
    MifOutputByteStream( MifDoc::T_indent i = 0 )
     : os( NULL ), CurTagIndent( i ) {};
    MifOutputByteStream( OutputByteStream &os, MifDoc::T_indent i = 0 )
     : os( &os ), CurTagIndent( i ) {};
    void setStream( OutputByteStream &s ) { os = &s; }
    OutputByteStream &stream() { assert( os != NULL ); return *os; }
    MifOutputByteStream &operator<<( unsigned n ) { stream() << n; return *this; }
    MifOutputByteStream &operator<<( long n ) { stream() << n; return *this; }
    MifOutputByteStream &operator<<( char c ) { stream() << c; return *this; }
    MifOutputByteStream &operator<<( const char *s ) { stream() << s; return *this; }
#ifdef SP_WIDE_SYSTEM
    MifOutputByteStream &operator<<( const String<CmdLineApp::AppChar> &s )
     { for( size_t i = 0; i < s.size(); i++ ) stream() << char( s[i] ); return *this; }
#endif
    MifOutputByteStream &operator<<( const String<char> &s ) { stream() << s; return *this; }
    MifOutputByteStream &operator<<( const MifDoc::T_string &s )
     { stream() << '`'; stream() << s; stream() << '\''; return *this; }
    MifOutputByteStream &operator<<( const MifDoc::T_dimension &d ) {
        char buf[32];
        int i;
        sprintf( buf, "%li.%.3i%n", long(d)/1000, abs(long(d)%1000), &i );
        while( buf[--i] == '0' ) {}; if( buf[i] == '.' ) i--;
        buf[i+1] = '\0';
        stream() << buf << "pt";
        return *this;
    }
    MifOutputByteStream &operator<<( const MifDoc::PotentialMarkerInfo &pmi ) {
        stream() << MifDoc::escapeChar();
        stream() << MifDoc::PotentialMarkerInfo::TypeChar;
        stream().sputn( (char*)&pmi, sizeof( pmi ) );
        return *this;
    }
    MifOutputByteStream &operator<<( const MifDoc::XRefInfo &xri ) {
        stream() << MifDoc::escapeChar();
        stream() << MifDoc::XRefInfo::TypeChar;
        stream().sputn( (char*)&xri, sizeof( xri ) );
        return *this;
    }
    MifOutputByteStream &operator<<( const MifDoc::T_LTRB &b )
     { *this << b.l << " " << b.t << " " << b.r << " " << b.b; return *this; }
    MifOutputByteStream &operator<<( const MifDoc::T_LTWH &b )
     { *this << b.l << " " << b.t << " " << b.w << " " << b.h; return *this; }
    MifOutputByteStream &operator<<( const MifDoc::T_WH &s )
     { *this << s.w << " " << s.h; return *this; }
    MifOutputByteStream &operator<<( const MifDoc::T_XY &s )
     { *this << s.x << " " << s.y; return *this; }
    MifOutputByteStream &operator<<( const MifDoc::T_boolean &b )
     { if( bool(b) ) *this << "Yes"; else *this << "No"; return *this; }
    MifOutputByteStream &operator<<( const MifDoc::T_indent &i )
     { unsigned cnt = (i==INDENT) ? CurTagIndent : unsigned( i );
       for( ; cnt; cnt-- ) stream() << ' '; return *this; }
};

class MifTmpOutputByteStream : public TmpOutputByteStream2 {

    MifOutputByteStream os;    

  public:
    MifTmpOutputByteStream( MifDoc::T_indent osIndent = 0 )
     : TmpOutputByteStream2(), os( osIndent ) { os.setStream( *this ); }
    void commit( OutputByteStream &os, bool resolveCrossReferences = false );
    void commit( String<char> &str );
    MifOutputByteStream  &stream() { return os; }
};

#ifdef DSSSL_NAMESPACE
}
#endif

#endif /* not MifDoc_INCLUDED */
