// Jade MIF backend
// Copyright (c) 1997 Kathleen Marszalek
// Version 1.0 (beta 2b)
// Date: October 26, 1997

#include "MifDoc.h"
#include "MifFOTBuilder.h"
#include "MifFOTBuilderPrivate.h"
#include "Message.h"
#include "InputSource.h"
#include "StorageManager.h"
#include "MifMessages.h"
#include "CharMap.h"

#include <assert.h>
#include <stdio.h>

#ifdef DSSSL_NAMESPACE
namespace DSSSL_NAMESPACE {
#endif

MifFOTBuilder *MifFOTBuilder::CurInstance = NULL;
unsigned MifFOTBuilder::NodeInfo::nonEmptyElementsOpened = 0;
unsigned MifFOTBuilder::NodeInfo::curNodeLevel = 0;

int operator ==(struct FOTBuilder::DisplayNIC, struct FOTBuilder::DisplayNIC) { return 0; }

void MifFOTBuilder::start() {

    NodeInfo::nonEmptyElementsOpened = nodeStack.size();

    Format *effectiveFormat = new Format( nextFormat );
    if( nextFormat.FSize > 0 )
        effectiveFormat->setFDY
         ( computeLengthSpec( nextFormat.FotPositionPointShiftSpec ) * -100 / nextFormat.FSize );

    formatStack.insert( effectiveFormat );
}

void MifFOTBuilder::end() {

    assert( formatStack.head() != NULL );
    delete formatStack.get();

    assert( formatStack.head() != NULL );
    nextFormat = *formatStack.head();
}

void MifFOTBuilder::Format::computePgfLanguage() {

    switch( FotLanguage ) {
      case SP_LETTER2( 'E', 'N' ):
        switch( FotCountry ) {
          case SP_LETTER2( 'G', 'B' ):
            PgfLanguage = MifDoc::sUKEnglish;
            break;
          default:
            PgfLanguage = MifDoc::sUSEnglish;
        }
        break;
      case SP_LETTER2( 'D', 'E' ):
        switch( FotCountry ) {
          case SP_LETTER2( 'C', 'H' ):
            PgfLanguage = MifDoc::sSwissGerman;
            break;
          default:
            PgfLanguage = MifDoc::sGerman;
        }
      case SP_LETTER2( 'F', 'R' ):
        switch( FotCountry ) {
          case SP_LETTER2( 'C', 'A' ):
            PgfLanguage = MifDoc::sCanadianFrench;
            break;
          default:
            PgfLanguage = MifDoc::sFrench;
        }
        break;
      case SP_LETTER2( 'E', 'S' ):
        PgfLanguage = MifDoc::sSpanish;
        break;
      case SP_LETTER2( 'C', 'A' ):
        PgfLanguage = MifDoc::sCatalan;
        break;
      case SP_LETTER2( 'I', 'T' ):
        PgfLanguage = MifDoc::sItalian;
        break;
      case SP_LETTER2( 'P', 'T' ):
        PgfLanguage = MifDoc::sPortuguese;
        break;
//      case SP_LETTER2( '', '' ):
//        PgfLanguage = sBrazilian;
//        break;
//      case SP_LETTER2( '', '' ):
//        PgfLanguage = sDanish;
//        break;
      case SP_LETTER2( 'N', 'L' ):
        PgfLanguage = MifDoc::sDutch;
        break;
      case SP_LETTER2( 'N', 'O' ):
        PgfLanguage = MifDoc::sNorwegian;
        break;
 //     case SP_LETTER2( '', '' ):
 //       PgfLanguage = sNynorsk;
 //       break;
      case SP_LETTER2( 'F', 'I' ):
        PgfLanguage = MifDoc::sFinnish;
        break;
      case SP_LETTER2( 'S', 'V' ):
        PgfLanguage = MifDoc::sSwedish;
        break;
      default:
        PgfLanguage = MifDoc::sNoLanguage;
    }
}

MifFOTBuilder::MifFOTBuilder( const String<CmdLineApp::AppChar> &fileLoc,
                              const Ptr<ExtendEntityManager> &entityManager,
 		            	      const CharsetInfo &systemCharset,
                              CmdLineApp *app )
 : App( app ), EntityManager( entityManager ), SystemCharset( &systemCharset ),
   mifDoc( fileLoc, app ), paragraphBreakInEffect( false ), inLeader( false ),
   CurLeaderStream( NULL ), CharTable( 0 ), lastFlowObjectWasWhitespace( false ),
   pendingBreak( symbolFalse ), firstHeaderFooter( true ),  inSimplePageSequence( false ),
   bookComponentOpened( false ), bookComponentAvailable( false ) {

    CurInstance = this;

    nextFormat.setDSSSLDefaults();
    nextFormat.FotCurDisplaySize = Format::INITIAL_PAGE_SIZE() - 2;
     // 2 is for margins (MIF doesn't accept zero-sized ones)
    formatStack.insert( new Format( nextFormat ) );

    initMifBookComponent();
    bookComponentOpened = true;
    bookComponentAvailable = true;

    // symbol character code from James Clark
    for( int i = 0; i < nSymbolFonts; i++ ) {
        for( int j = 0; j < 256; j++ ) {
            Char c = SymbolFonts[i].mapping[j];
            if( c && !CharTable[c] )
            	CharTable.setChar( c, j | (i << CHAR_TABLE_CHAR_BITS) | CHAR_TABLE_SYMBOL_FLAG );
        }
        StringC tem;
        for( const char *s = SymbolFonts[i].name; *s; s++ )
            tem += *s;
    }
}

void MifFOTBuilder::initMifBookComponent() {

   MifDoc::ParagraphFormat defaultParaFormat;
   defaultParaFormat.setDSSSLDefaults();
   defaultParaFormat.setPgfTag( MifDoc::sDefaultPgfFormat );
   mifDoc.pgfCatalog().ParaFormats.push_back( defaultParaFormat );
 
   mifDoc.document().setDPageSize
    ( MifDoc::T_WH( MifDoc::T_dimension( format().FotPageWidth ),
                    MifDoc::T_dimension( format().FotPageHeight ) ) );
   mifDoc.document().setDMargins( MifDoc::T_LTRB( 1, 1, 1, 1 ) );
    // MIF doesn't accept zeros
   mifDoc.document().setDColumns( 1 );
}

long MifFOTBuilder::computeLengthSpec( const LengthSpec &spec ) const {

    if( spec.displaySizeFactor == 0.0 ) {
        return spec.length;
    } else {
        double tem = format().FotCurDisplaySize * spec.displaySizeFactor;
        return spec.length + long( tem >= 0.0 ? tem +.5 : tem - .5 );
    }
}

MifFOTBuilder::EffectiveDisplaySpace
MifFOTBuilder::createEffectiveDisplaySpace( const DisplaySpace &ds ) const {

    EffectiveDisplaySpace result;
    result.set( computeLengthSpec( ds.nominal ), computeLengthSpec( ds.min ),
                computeLengthSpec( ds.max ), ds.priority, ds.conditional, ds.force );
    return result;
}

void MifFOTBuilder::EffectiveDisplaySpace::combine( const EffectiveDisplaySpace eds ) {

    if( eds.force ) {
        if( force ) {
            nominal += eds.nominal;
            min += eds.min;
            max += eds.max;
        } else
            *this = eds;
    } else {
        if( eds.priority > priority )
            *this = eds;
        else {
            if( eds.priority == priority ) {
                if( eds.nominal > nominal )
                    *this = eds;
                else
                    if( eds.nominal == nominal ) {
                        if( eds.min < min ) {
                            min = eds.min;
                        }
                        if( eds.max > max ) {
                            max = eds.max;
                        }
                    }
            }
        }
    }
}

void MifFOTBuilder::synchronizeFontFormat() {

    if( mifDoc.curPara( false ) != NULL ) {
//        mifDoc.curPara()->curFormat().ffUpdateFrom( format() );
        mifDoc.curPara()->curFormat().ffUpdateFrom( nextFormat );
        mifDoc.curPara()->curFormat().ffOut( mifDoc.os(), MifDoc::FontFormat::stFont );
    } else {
//        mifDoc.curFormat().ffUpdateFrom( format() );
        mifDoc.curFormat().ffUpdateFrom( nextFormat );
        mifDoc.curFormat().ffOut( mifDoc.os(), MifDoc::FontFormat::stFont );
    }
    outPotentialMarkers();
}

void MifFOTBuilder::outString
 ( const Char *s, size_t n, MifTmpOutputByteStream *o, bool inParagraph,
   String<char> *targetString ) {

    MifOutputByteStream *outS = ( o != NULL ) ? &o->stream()
                                              : ( ( targetString != NULL ) ? NULL : &mifDoc.os() );

    enum { nStr, nCode, nChar, nNothing, nHardReturn, nSpecialChar } nextOut;
    union { const char *str; unsigned long code; char chr;  } out;
    String<char> out_specialChar;

    MifDoc::ParagraphFormat *curPFormat
     = inParagraph 
        ? ( mifDoc.curPara( false ) != NULL ? &mifDoc.curPara()->curFormat() : &mifDoc.curFormat() ) 
        : (MifDoc::ParagraphFormat*)NULL;
    MifDoc::T_string paraFFamily;
    if( curPFormat != NULL )
        paraFFamily = curPFormat->FFamily;
    bool stringOpened = false;
    int prevSymbolFontNum = 0;
    int curSymbolFontNum = 0;
    bool thisFlowObjectIsWhitespace;
    for( ; n > 0; s++, n-- ) {
        thisFlowObjectIsWhitespace = false;
        nextOut = nNothing;
        switch( *s ) {
          case Char( '\n' ): break;
          case Char( '\r' ):
            if( !inParagraph ) {
                out.chr = ' '; nextOut = nChar;
            } else {
                switch( format().FotLines ) {
                  case symbolNone:
                  case symbolWrap:
                    switch( format().FotInputWhitespaceTreatment ) {
                      case symbolIgnore: break;
                      case symbolCollapse:
                        if( lastFlowObjectWasWhitespace )
                            break;
                      case symbolPreserve: default:            
                        out.chr = ' '; nextOut = nChar;
                    }  
                    break;
                  default:
                    nextOut = nHardReturn;
                }
                thisFlowObjectIsWhitespace = true;
            }
            break;
          case Char( '\t' ):
            if( !inParagraph ) {
                out.str = "\\t"; nextOut = nStr;
            } else {
                switch( format().FotInputWhitespaceTreatment ) {
                  case symbolIgnore: break;
                  case symbolCollapse:
                    if( lastFlowObjectWasWhitespace )
                        break;
                  case symbolPreserve: default:            
                    out.str = "\\t"; nextOut = nStr;
                }  
                thisFlowObjectIsWhitespace = true;
            }
            break;
          case Char( '>' ):  out.str = "\\>"; nextOut = nStr; break;
          case Char( '\'' ): out.str = "\\q"; nextOut = nStr; break;
          case Char( '`' ):  out.str = "\\Q"; nextOut = nStr; break;
          case Char( '\\' ): out.str = "\\\\"; nextOut = nStr; break;
          case Char( ' ' ):
            if( !inParagraph ) {
                out.chr = ' '; nextOut = nChar;
            } else {
                switch( format().FotInputWhitespaceTreatment ) {
                  case symbolIgnore: break;
                  case symbolCollapse:
                    if( lastFlowObjectWasWhitespace )
                        break;
                  case symbolPreserve: default:            
                    out.chr = ' '; nextOut = nChar;
                }  
                thisFlowObjectIsWhitespace = true;
            }
            break;
          case 0x00A0:
            out_specialChar = MifDoc::sHardSpace; nextOut = nSpecialChar;
            break;
          case 0x00A2:
            out_specialChar = MifDoc::sCent; nextOut = nSpecialChar;
            break;
          case 0x00A3:
          case 0x20A4:
            out_specialChar = MifDoc::sPound; nextOut = nSpecialChar;
            break;
          case 0x00A5:
            out_specialChar = MifDoc::sYen; nextOut = nSpecialChar;
            break;
          case 0x2002:
            out_specialChar = MifDoc::sEnSpace; nextOut = nSpecialChar;
            break;
          case 0x2003:
            out_specialChar = MifDoc::sEmSpace; nextOut = nSpecialChar;
            break;
          case 0x2009:
            out_specialChar = MifDoc::sThinSpace; nextOut = nSpecialChar;
            break;
          case 0x2010:
            out_specialChar = MifDoc::sSoftHyphen; nextOut = nSpecialChar;
            break;
          case 0x2011:
            out_specialChar = MifDoc::sHardHyphen; nextOut = nSpecialChar;
            break;
          case 0x2013:
            out_specialChar = MifDoc::sEnDash; nextOut = nSpecialChar;
            break;
          case 0x2014:
            out_specialChar = MifDoc::sEmDash; nextOut = nSpecialChar;
            break;
          case 0x2020:
            out_specialChar = MifDoc::sDagger; nextOut = nSpecialChar;
            break;
          case 0x2021:
            out_specialChar = MifDoc::sDoubleDagger; nextOut = nSpecialChar;
            break;
          case 0x2022:
            out_specialChar = MifDoc::sBullet; nextOut = nSpecialChar;
            break;
          default:
            if( *s >= 0x80 ) {
                unsigned long code = CharTable[*s];
	            if( code & CHAR_TABLE_SYMBOL_FLAG ) {
                    if( curPFormat != NULL ) {
                        curSymbolFontNum = ( ( code & ~CHAR_TABLE_SYMBOL_FLAG )
                                              >> CHAR_TABLE_CHAR_BITS ) + 1;
                        assert( curSymbolFontNum < nSymbolFonts );
                        out.code = code & 0xff; nextOut = nCode;
                    }
                } else
                    if( code ) {
                        out.code = code & 0xff; nextOut = nCode;
                    }
            } else {
                out.chr = char( *s ); nextOut = nChar;
            }
        }

        if( nextOut != nNothing ) {
            if( curSymbolFontNum != prevSymbolFontNum ) {
                if( stringOpened ) {
                    *outS << "'>";
                    stringOpened = false;
                }
                MifDoc::T_string fontFamily( curSymbolFontNum == 0 
                                      ? paraFFamily
                                      : MifDoc::T_string( 
                                         SymbolFonts[ curSymbolFontNum - 1 ].name ) );
                curPFormat->setFFamily( fontFamily );
                curPFormat->ffOut( *outS, MifDoc::FontFormat::stFont );
            }
            if( curPFormat != NULL && !stringOpened ) {
                *outS << '\n' << MifOutputByteStream::INDENT << "<String `";
                stringOpened = true;
            }
            switch( nextOut ) {
              case nChar:
                  if( targetString != NULL )
                    *targetString += out.chr;
                  else
                    *outS << out.chr;
                  break;

              case nHardReturn:
                if( inParagraph ) {
                    if( stringOpened ) {
                        *outS << "'>";
                        stringOpened = false;
                    }
                    mifDoc.outSpecialChar( MifDoc::sHardReturn, outS );
                    MifDoc::ParaLine::outEpilog( *outS );
                    MifDoc::ParaLine::outProlog( *outS );
                }
                break;

              case nSpecialChar:
                if( inParagraph ) {
                    if( stringOpened ) {
                        *outS << "'>";
                        stringOpened = false;
                    }
                    mifDoc.outSpecialChar( out_specialChar, outS );
                }
                break;

              case nCode:
                if( targetString != NULL )
                   mifDoc.outHexChar( out.code, *targetString );
                else
                   mifDoc.outHexChar( out.code, outS ); break;
                break;          

              case nStr:
                if( targetString != NULL )
                    targetString->append( out.str, strlen( out.str ) );
                else
                    *outS << out.str;
                break;
              case nNothing: break;
            }
        }

        prevSymbolFontNum = curSymbolFontNum;    
        lastFlowObjectWasWhitespace = thisFlowObjectIsWhitespace;
    }

    if( stringOpened )
        *outS << "'>";

    if( curSymbolFontNum != 0 ) {
        curPFormat->setFFamily( paraFFamily );
        curPFormat->ffOut( *outS, MifDoc::FontFormat::stFont );
        curPFormat->ffSetProperties() &= ~MifDoc::FontFormat::fFFamily;
    }
}

void MifFOTBuilder::characters( const Char *s, size_t n ) {

    checkForParagraphReopening();
    if( MifDoc::Para::currentlyOpened ) {
        if( inLeader ) {
            outString( s, n, &curLeaderStream(), false );
        } else {
            synchronizeFontFormat();
            outString( s, n, NULL, true );
        }
    }
}

void MifFOTBuilder::checkForParagraphReopening() {

   DisplayInfo *curDs = displayStack.head();
//   assert( curDs != NULL );
   if( curDs != NULL && curDs->paragraphClosedInMif ) {

        Format f = format();

        f.setPgfWithPrev( curDs->firstParaOutputed ? curDs->keepWithinPageInEffect : false );
        f.setPgfWithNext( false ); // TMAI
        curDs->firstParaOutputed = true;

        processDisplaySpaceStack();
        switch( pendingBreak ) {
            case symbolPage:
                f.setPgfPlacement( MifDoc::sPageTop );
                break;
            case symbolColumn:
                f.setPgfPlacement( MifDoc::sColumnTop );
                break;
            default:
                f.setPgfPlacement( MifDoc::sAnywhere );
        }
        pendingBreak = symbolFalse;

        f.setPgfSpBefore( pendingEffectiveDisplaySpace.nominal );
         pendingEffectiveDisplaySpace.clear();
        f.setPgfFIndent( nextFormat.PgfLIndent );

        MifDoc::Para *p
         = new MifDoc::Para( mifDoc.curTagStream().content().stream().CurTagIndent + 4 );
        p->setParagraphFormat( mifDoc.curFormat() );
        p->format().updateFrom( f );
        p->format().FSize = mifDoc.curFormat().FSize;
        p->format().ffSetProperties() &= ~MifDoc::FontFormat::fFSize;
        Length lineSpacing = computeLengthSpec( f.FotLineSpacingSpec.length );
        if( p->format().FSize != lineSpacing )
            p->format().setFSize( lineSpacing );
        p->curFormat() = p->format();
        // assert( mifDoc.curPara( false ) == NULL );
        mifDoc.setCurPara( p );
        mifDoc.enterPara( *mifDoc.curPara() );

        curDs->paragraphClosedInMif = false;
   }
}

void MifFOTBuilder::processDisplaySpaceStack() {

    pendingBreak = symbolFalse;
    EffectiveDisplaySpace effectiveDisplaySpace;
    DisplaySpaceInfo *curDSI;
    while( !displaySpaceQueue.empty() ) {
        curDSI = displaySpaceQueue.get();
        if( curDSI->breakType == symbolPage || curDSI->breakType == symbolColumn ) {
            effectiveDisplaySpace.clear();
            if( pendingBreak != symbolFalse )
                mifDoc.outBreakingPara( curDSI->breakType == symbolPage
                                         ? MifDoc::sPageTop : MifDoc::sColumnTop );
            pendingBreak = curDSI->breakType;
            if( !curDSI->breakIsAfter )
                effectiveDisplaySpace.combine( createEffectiveDisplaySpace( curDSI->space ) );
        } else
            effectiveDisplaySpace.combine( createEffectiveDisplaySpace( curDSI->space ) );
        delete curDSI;
    }

    pendingEffectiveDisplaySpace = effectiveDisplaySpace;
}


void MifFOTBuilder::doStartParagraph
      ( const ParagraphNIC &nic, bool servesAsWrapper, MifDoc::T_dimension height,
        bool allowNegativeLeading ) {

    startDisplay( nic );

    DisplayInfo *curDs = displayStack.head();
    assert( curDs != NULL );
    curDs->isParagraph = true;

    nextFormat.setPgfWithPrev( curDs->mayViolateKeepBefore
                                ? curDs->keepWithPrevious
                                : ( curDs->firstParaOutputed
                                     ? curDs->keepWithinPageInEffect
                                     : false ) );
    nextFormat.setPgfWithNext( curDs->keepWithNext ); 
    curDs->firstParaOutputed = true;

    processDisplaySpaceStack();
    switch( pendingBreak ) {
        case symbolPage:
            nextFormat.setPgfPlacement( MifDoc::sPageTop );
            break;
        case symbolColumn:
            nextFormat.setPgfPlacement( MifDoc::sColumnTop );
            break;
        default:
            nextFormat.setPgfPlacement( MifDoc::sAnywhere );
    }
    pendingBreak = symbolFalse;

    Length lineSpacing;
    Length extraSpaceBefore = 0; 
    if( servesAsWrapper ) {
        if( height < 2000 ) {   // FrameMaker minimum is 2pt    
            if( allowNegativeLeading )
                nextFormat.setPgfLeading( height - 2000 );
            else
                pendingEffectiveDisplaySpace.nominal
                  -= ( pendingEffectiveDisplaySpace.nominal > 2000 )
                        ? 2000 : pendingEffectiveDisplaySpace.nominal;
                 // try to steal as much as possible from space before
            height = 2000;
        }
        //if( effectiveDisplaySpace.nominal <= 0 )
         //    extraSpaceBefore = 1; // FrameMaker bug workaround    
        //nextFormat.setFSize( height );
        lineSpacing = height;
        nextFormat.setFColor( MifDoc::sWhite );
        nextFormat.setPgfLineSpacing( MifDoc::sFixed );
    } else {
        //nextFormat.setFSize( computeLengthSpec( nextFormat.FotLineSpacingSpec.length ) );
        lineSpacing = computeLengthSpec( nextFormat.FotLineSpacingSpec.length );
        if( nextFormat.FotMinLeading.hasLength ) // but ignore the actual min-leading value
            nextFormat.PgfLineSpacing = MifDoc::sProportional;
        else
            nextFormat.PgfLineSpacing = MifDoc::sFixed;
    }
    
    nextFormat.setPgfSpBefore( pendingEffectiveDisplaySpace.nominal + extraSpaceBefore );
    pendingEffectiveDisplaySpace.clear();

    nextFormat.setPgfLIndent( computeLengthSpec( nextFormat.FotStartIndentSpec ) );
    nextFormat.setPgfFIndent( computeLengthSpec( nextFormat.FotFirstLineStartIndentSpec )
                               + nextFormat.PgfLIndent );
    nextFormat.setPgfRIndent( computeLengthSpec( nextFormat.FotEndIndentSpec ) );

    nextFormat.setPgfPlacementStyle
     ( nextFormat.FotSpan > 1 ? MifDoc::sStraddle : MifDoc::sNormal );
    start();

    MifDoc::Para *p
     = new MifDoc::Para( mifDoc.curTagStream().content().stream().CurTagIndent + 4 );
    p->setParagraphFormat( mifDoc.curFormat() );
    p->format().updateFrom( format() );
    p->format().FSize = mifDoc.curFormat().FSize;
    p->format().ffSetProperties() &= ~MifDoc::FontFormat::fFSize;
    if( p->format().FSize != lineSpacing )
        p->format().setFSize( lineSpacing );
    p->curFormat() = p->format();
    // assert( mifDoc.curPara( false ) == NULL );
    mifDoc.setCurPara( p );
    mifDoc.enterPara( *mifDoc.curPara() );
    lastFlowObjectWasWhitespace = false;
    outPotentialMarkers();

    // mifDoc.curFormat().out( mifDoc.os() );
    // mifDoc.beginParaLine();
}

void MifFOTBuilder::doEndParagraph
 ( bool sustainFormatStack, bool sustainDisplayStack, bool paragraphBreakTest,
   bool discardThisPara ) {

//    mifDoc.endParaLine();
//    MifDoc::Para::outEpilog( mifDoc.os() );

    DisplayInfo *curDs = displayStack.head();
    assert( curDs != NULL );

    if( !sustainFormatStack )
        end();
    if( !sustainDisplayStack )
        endDisplay();

    if( paragraphBreakTest && paragraphBreakInEffect ) {
        paragraphBreakInEffect = false;
        end();
    }

    if( !curDs->paragraphClosedInMif ) {
        MifDoc::Para *p = mifDoc.curPara();
        mifDoc.exitPara();
        if( !discardThisPara ) {
            p->out( mifDoc.os() );
            mifDoc.curFormat().updateFrom( p->format() );
        }
        delete p;
    }
}

void MifFOTBuilder::startDisplayGroup( const DisplayGroupNIC &nic ) {

    startDisplay( nic );
    start();
}

void MifFOTBuilder::endDisplayGroup() {

    end();
    endDisplay();
}

void MifFOTBuilder::startDisplay( const DisplayNIC &nic ) {

    if( !inSimplePageSequence ) {
        if( !bookComponentOpened ) {
            mifDoc.enterBookComponent();
            initMifBookComponent();
            bookComponentOpened = true;
        }
        bookComponentAvailable = false;
    }

    displaySpaceQueue.append
     ( new DisplaySpaceInfo( nic.spaceBefore, nic.breakBefore, false ) );

    DisplayInfo *di = displayStack.head();
    if( di != NULL && di->isParagraph && !di->paragraphClosedInMif ) {
        if( mifDoc.curPara()->content().isEmpty() ) {
            doEndParagraph( true, true, true, true );
        } else
            doEndParagraph( true, true );
        di->paragraphClosedInMif = true;
    }

    if( nextFormat.FotSpan > 1 )
        nextFormat.FotCurDisplaySize
         = nextFormat.FotPageWidth - nextFormat.FotLeftMargin - nextFormat.FotRightMargin;
    else
        nextFormat.FotCurDisplaySize
         = ( nextFormat.FotPageWidth - nextFormat.FotLeftMargin - nextFormat.FotRightMargin
              - nextFormat.FotPageColumnSep * ( nextFormat.FotPageNColumns - 1 ) )  
            / nextFormat.FotPageNColumns;

    displayStack.insert( new DisplayInfo( nic, displayStack.head() ) );
}

void MifFOTBuilder::endDisplay() {

    DisplayInfo *di = displayStack.get();
    
    displaySpaceQueue.append
     ( new DisplaySpaceInfo( di->spaceAfter, di->breakAfter, true ) );

    delete di;
}

void MifFOTBuilder::startSimplePageSequence() {

    inSimplePageSequence = true;
    firstHeaderFooter = true;

    bool openBookComponent = true;
    if( bookComponentOpened ) {
        if( bookComponentAvailable ) {
            openBookComponent = false;
            bookComponentAvailable = false;
        } else
            mifDoc.exitBookComponent();
    }

    if( openBookComponent ) {
        mifDoc.enterBookComponent();
        initMifBookComponent();
        bookComponentOpened = true;
        bookComponentAvailable = false;
    }

    nextFormat.FotCurDisplaySize
     = ( nextFormat.FotPageWidth - nextFormat.FotLeftMargin - nextFormat.FotRightMargin
          - nextFormat.FotPageColumnSep * ( nextFormat.FotPageNColumns - 1 ) )  
        / nextFormat.FotPageNColumns;

    mifDoc.document().setDColumns( nextFormat.FotPageNColumns );
    if( nextFormat.FotPageNColumns > 1 )
        mifDoc.document().setDColumnGap( nextFormat.FotPageColumnSep );

    mifDoc.document().setProperties &= ~MifDoc::Document::fDMargins;

    start();

    FotSimplePageSequence.paragraphFormat = format();
}

void MifFOTBuilder::endSimplePageSequence() {

    end();
    mifDoc.exitTextFlow();
    mifDoc.exitBookComponent();
    inSimplePageSequence = false;
    bookComponentOpened = false;
    bookComponentAvailable = false;
}

void MifFOTBuilder::makeEmptyTextFlow( MifDoc::TextRect &textRect ) {

    MifDoc::TextFlow &textFlow = *(new MifDoc::TextFlow( textRect, true ));
    mifDoc.textFlows().push_back( textFlow );

    mifDoc.enterTextFlow( textFlow );
    MifDoc::Para::outSimpleProlog( mifDoc.os() );
    MifDoc::ParaLine::outProlog( mifDoc.os() );
    MifDoc::ParaLine::outEpilog( mifDoc.os() );
    MifDoc::Para::outEpilog( mifDoc.os() );
    mifDoc.exitTextFlow();
}

void MifFOTBuilder::beginHeaderFooter( bool header ) {

    start();

    MifDoc::Para p;
    p.setPgfTag( header ? MifDoc::sHeader : MifDoc::sFooter );
    p.setProperties &= ~MifDoc::Para::fParagraphFormat;
    p.outProlog( mifDoc.os() );

    MifDoc::FontFormat ff;
    ff.setFSize( format().FSize );
    ff.out( mifDoc.os(), MifDoc::FontFormat::fFSize, MifDoc::FontFormat::stFont );

    mifDoc.beginParaLine();
}

void MifFOTBuilder::endHeaderFooter() {

    mifDoc.endParaLine();
    MifDoc::Para::outEpilog( mifDoc.os() );

    end();
}

void MifFOTBuilder::setupSimplePageSequence() {

    MifDoc::Page &firstMasterPage
     = *(new MifDoc::Page( MifDoc::sOtherMasterPage, MifDoc::sFirst ));
    MifDoc::Page &rightMasterPage
     = *(new MifDoc::Page( MifDoc::sRightMasterPage, MifDoc::sRight ));
    MifDoc::Page &leftMasterPage
     = *(new MifDoc::Page( MifDoc::sLeftMasterPage, MifDoc::sLeft ));
    MifDoc::Page &bodyPage
     = *(new MifDoc::Page( MifDoc::sBodyPage, MifDoc::sNONE, MifDoc::sFirst ));

    MifDoc::T_LTWH bodyRect;
    MifDoc::T_LTWH headerRect;
    MifDoc::T_LTWH footerRect;

    bodyRect.l = format().FotLeftMargin;
    bodyRect.t = format().FotTopMargin;
    bodyRect.w = format().FotPageWidth - format().FotLeftMargin - format().FotRightMargin;
    bodyRect.h = format().FotPageHeight - format().FotTopMargin - format().FotBottomMargin;

    headerRect.l = format().FotLeftMargin;
    headerRect.t = 0;
    headerRect.w = bodyRect.w;
    headerRect.h = format().FotTopMargin;

    footerRect.l = format().FotLeftMargin;
    footerRect.t = format().FotPageHeight - format().FotBottomMargin;

    footerRect.w = bodyRect.w;
    footerRect.h = format().FotBottomMargin;
   {
    MifDoc::TextRect &firstBodyTextRect
     = *(new MifDoc::TextRect( bodyRect, format().FotPageNColumns, format().FotPageColumnSep,
                               format().FotPageBalanceColumns ));
    MifDoc::TextRect &rightBodyTextRect
     = *(new MifDoc::TextRect( bodyRect, format().FotPageNColumns, format().FotPageColumnSep,
                               format().FotPageBalanceColumns));
    MifDoc::TextRect &leftBodyTextRect
     = *(new MifDoc::TextRect( bodyRect, format().FotPageNColumns, format().FotPageColumnSep,
                               format().FotPageBalanceColumns ));
    MifDoc::TextRect &bodyTextRect
     = *(new MifDoc::TextRect( bodyRect, format().FotPageNColumns, format().FotPageColumnSep,
                               format().FotPageBalanceColumns ));
    MifDoc::TextRect &firstHeaderTextRect = *(new MifDoc::TextRect( headerRect ));
    MifDoc::TextRect &rightHeaderTextRect = *(new MifDoc::TextRect( headerRect ));
    MifDoc::TextRect &leftHeaderTextRect = *(new MifDoc::TextRect( headerRect ));
    MifDoc::TextRect &firstFooterTextRect = *(new MifDoc::TextRect( footerRect ));
    MifDoc::TextRect &rightFooterTextRect = *(new MifDoc::TextRect( footerRect ));
    MifDoc::TextRect &leftFooterTextRect = *(new MifDoc::TextRect( footerRect ));

    firstMasterPage.TextRects.push_back( firstHeaderTextRect );
    firstMasterPage.TextRects.push_back( firstBodyTextRect );
    firstMasterPage.TextRects.push_back( firstFooterTextRect );

    rightMasterPage.TextRects.push_back( rightHeaderTextRect );
    rightMasterPage.TextRects.push_back( rightBodyTextRect );
    rightMasterPage.TextRects.push_back( rightFooterTextRect );

    leftMasterPage.TextRects.push_back( leftHeaderTextRect );
    leftMasterPage.TextRects.push_back( leftBodyTextRect );
    leftMasterPage.TextRects.push_back( leftFooterTextRect );

    bodyPage.TextRects.push_back( bodyTextRect );
    mifDoc.pages().push_back( bodyPage );
    mifDoc.pages().push_back( firstMasterPage );
    mifDoc.pages().push_back( rightMasterPage );
    mifDoc.pages().push_back( leftMasterPage );

    MifDoc::ParagraphFormat headerPF;
    headerPF.setDSSSLDefaults();
    MifDoc::ParagraphFormat footerPF;
    footerPF.setDSSSLDefaults();
    setupHeaderFooterParagraphFormat( headerPF, footerPF, bodyRect.w );

    FotSimplePageSequence.BodyTextFlow
     = new MifDoc::TextFlow( bodyTextRect, true, &FotSimplePageSequence.paragraphFormat,
                             MifDoc::sDefaultPgfFormat );
    FotSimplePageSequence.FirstHeaderTextFlow
     = new MifDoc::TextFlow( firstHeaderTextRect, false, &headerPF, MifDoc::sHeader );
    FotSimplePageSequence.FirstFooterTextFlow
     = new MifDoc::TextFlow( firstFooterTextRect, false, &footerPF, MifDoc::sFooter );
    FotSimplePageSequence.LeftHeaderTextFlow
     = new MifDoc::TextFlow( leftHeaderTextRect, false, &headerPF, MifDoc::sHeader );
    FotSimplePageSequence.LeftFooterTextFlow
     = new MifDoc::TextFlow( leftFooterTextRect, false, &footerPF, MifDoc::sFooter );
    FotSimplePageSequence.RightHeaderTextFlow
     = new MifDoc::TextFlow( rightHeaderTextRect, false, &headerPF, MifDoc::sHeader );
    FotSimplePageSequence.RightFooterTextFlow
     = new MifDoc::TextFlow( rightFooterTextRect, false, &footerPF, MifDoc::sFooter );

    makeEmptyTextFlow( firstBodyTextRect );
    makeEmptyTextFlow( leftBodyTextRect );
    makeEmptyTextFlow( rightBodyTextRect );

    mifDoc.textFlows().push_back( *FotSimplePageSequence.BodyTextFlow );
    mifDoc.textFlows().push_back( *FotSimplePageSequence.FirstHeaderTextFlow );
    mifDoc.textFlows().push_back( *FotSimplePageSequence.FirstFooterTextFlow );
    mifDoc.textFlows().push_back( *FotSimplePageSequence.LeftHeaderTextFlow );
    mifDoc.textFlows().push_back( *FotSimplePageSequence.LeftFooterTextFlow );
    mifDoc.textFlows().push_back( *FotSimplePageSequence.RightHeaderTextFlow );
    mifDoc.textFlows().push_back( *FotSimplePageSequence.RightFooterTextFlow );

    mifDoc.document().setDTwoSides( true );
    mifDoc.document().setDParity( MifDoc::sFirstRight );
  }

  mifDoc.enterTextFlow( *FotSimplePageSequence.BodyTextFlow );
}

void MifFOTBuilder::startSimplePageSequenceHeaderFooter( unsigned hfPart ) {

    if( firstHeaderFooter ) { setupSimplePageSequence(); firstHeaderFooter = false; }

    MifDoc::TextFlow *curTextFlow;

    if( hfPart & firstHF )
        if( hfPart & frontHF )
            if( hfPart & headerHF )
                curTextFlow = FotSimplePageSequence.FirstHeaderTextFlow;
            else
                curTextFlow = FotSimplePageSequence.FirstFooterTextFlow;
        else
            return;
    else
        if( hfPart & frontHF )
            if( hfPart & headerHF )
                curTextFlow = FotSimplePageSequence.RightHeaderTextFlow;
            else
                curTextFlow = FotSimplePageSequence.RightFooterTextFlow;
        else
            if( hfPart & headerHF )
                curTextFlow = FotSimplePageSequence.LeftHeaderTextFlow;
            else
                curTextFlow = FotSimplePageSequence.LeftFooterTextFlow;

    mifDoc.enterTextFlow( *curTextFlow );

    if( hfPart & (centerHF | rightHF) )
        mifDoc.outSpecialChar( MifDoc::sTab );
    else // leftHF
        (hfPart & headerHF) ? beginHeader() : beginFooter();
}

void MifFOTBuilder::endSimplePageSequenceHeaderFooter( unsigned hfPart ) {

    if( hfPart & rightHF && (hfPart & frontHF || !(hfPart & firstHF)) ) {
        endHeaderFooter();
    }

    if( !(hfPart & firstHF) || hfPart & frontHF )
        mifDoc.exitTextFlow();
}

void MifFOTBuilder::setupHeaderFooterParagraphFormat
 ( MifDoc::ParagraphFormat &hpf, MifDoc::ParagraphFormat &fpf,
   MifDoc::T_dimension textRectWidth ) {

    MifDoc::TabStop centerTS( MifDoc::sCenter, textRectWidth / 2 );
    MifDoc::TabStop rightTS( MifDoc::sRight, textRectWidth );

    hpf.setFrom( FotSimplePageSequence.paragraphFormat, 0, MifDoc::FontFormat::fAll );
    hpf.TabStops.push_back( centerTS );
    hpf.TabStops.push_back( rightTS );
    hpf.setProperties |= MifDoc::ParagraphFormat::fTabStops;
    fpf = hpf;
    hpf.setFSize( ( ( format().FotBottomMargin - format().FotFooterMargin ) * 3 ) / 2 );
    fpf.setFSize( ( format().FotHeaderMargin * 3 ) / 2 );
    hpf.setPgfTag( MifDoc::sHeader );
    fpf.setPgfTag( MifDoc::sFooter );

    mifDoc.pgfCatalog().ParaFormats.push_back( hpf );
    mifDoc.pgfCatalog().ParaFormats.push_back( fpf );

    hpf.clearSetProperties();
    fpf.clearSetProperties();
}

void MifFOTBuilder::endAllSimplePageSequenceHeaderFooter() {

//    mifDoc.enterTextFlow( *FotSimplePageSequence.BodyTextFlow );
}

inline void MifFOTBuilder::pageNumber() {

    mifDoc.outPageNumber();
}

// --------- MifFOTBuilder::Table ---------------------------------------------

void MifFOTBuilder::Table::resolveBorders
      ( Vector<Row> &rows, bool hasFirstTableRow, bool hasLastTableRow ) {

    bool isFirstRow;
    bool isLastRow;
    bool isFirstColumn;
    bool isLastColumn;
    Cell *cell = NULL;
    size_t r, c, rr, cc;
    bool leftEdge, topEdge;

    for( r = 0; r < rows.size(); r++ ) {
        for( c = 0; c < rows[r].Cells.size() - 1; c++ ) {
            cell = &rows[r].Cells[c];
            if( cell->OverlappingCell == NULL ) {
                for( rr = r, topEdge = true; rr < r + cell->nRowsSpanned; rr++ ) {
                    isFirstRow = ( rr == 0 ) ? true : false;
                    isLastRow = ( rr == rows.size() - 1 ) ? true : false;
                    for( cc = c, leftEdge = true; cc < c + cell->nColumnsSpanned; cc++ ) {
                        rows[rr].Cells[cc].OverlappingCell = cell;
                        isFirstColumn = ( cc == 0 ) ? true : false;
                        isLastColumn = ( cc == rows[rr].Cells.size() - 2 ) ? true : false;

                        if( leftEdge )
                            if( isFirstColumn ) {
                                cell->beforeColumnBorder.resolve( beforeColumnBorder );
                            } else
                                cell->beforeColumnBorder.resolve
                                 ( rows[rr].Cells[cc-1].OverlappingCell->afterColumnBorder );

                        if( topEdge )
                            if( isFirstRow && hasFirstTableRow )
                                cell->beforeRowBorder.resolve( beforeRowBorder );
                            else
                                if( !isFirstRow )
                                    cell->beforeRowBorder.resolve
                                     ( rows[rr-1].Cells[cc].OverlappingCell->afterRowBorder );

                        if( isLastColumn )
                            cell->afterColumnBorder.resolve( afterColumnBorder );

                        if( isLastRow && hasLastTableRow )
                            cell->afterRowBorder.resolve( afterRowBorder );
                        
                        leftEdge = false;
                    }
                    topEdge = false;
                }
            }                
        }
    }
}

void MifFOTBuilder::Table::begin( MifDoc &mifDoc ) {

    CurCell = NULL;
    NoTablePartsSeen = true;

    TableParts.resize( 0 );
    TableParts.resize( 1 );
    TableParts.back().begin( *this, mifDoc );
}

// --------- MifFOTBuilder::TablePart -----------------------------------------

void MifFOTBuilder::TablePart::begin( Table &parent, MifDoc &mifDoc ) {

    Columns.resize( 0 );
    Header.resize( 0 );
    Body.resize( 0 );
    Footer.resize( 0 );
    
    columnsProcessed = false;
    needsColumnReprocessing = false;

    ParentTable = &parent;
    parentTable().CurRows = &Body;    
    parentTable().CurTablePart = this;
    
    if( MifTableNum == 0 ) {
        mifDoc.tbls().resize( mifDoc.tbls().size() + 1 );
        MifTableNum = mifDoc.tbls().size();
        MifDoc::CurInstance->setCurTblNum( MifTableNum );
    }
}

void MifFOTBuilder::TablePart::translate( MifDoc &mifDoc ) {

    if( needsColumnReprocessing )
        processColumns();

    MifDoc::Tbl &mifTbl = mifTable( mifDoc );
    if( parentTable().startIndent != 0 ) { // DSSSL default
        mifTbl.TblFormat.setTblLIndent( parentTable().startIndent );
        mifTbl.setProperties |= MifDoc::Tbl::fTblFormat;
    }

    if( parentTable().displayAlignment != symbolStart ) {
        mifTbl.setProperties |= MifDoc::Tbl::fTblFormat;
        MifDoc::T_keyword mifAlignment = MifDoc::sLeft;
        switch( parentTable().displayAlignment ) {
            case symbolStart:   mifAlignment = MifDoc::sLeft; break;
            case symbolEnd:     mifAlignment = MifDoc::sRight; break;
            case symbolCenter:  mifAlignment = MifDoc::sCenter; break;
            case symbolInside:  mifAlignment = MifDoc::sInside; break;
            case symbolOutside: mifAlignment = MifDoc::sOutside; break;
            default: assert( false );
        }
        mifTbl.TblFormat.setTblAlignment( mifAlignment );
    }

    bool putHeaderInBody = Body.size() == 0 && Header.size() > 0;
    bool putFooterInBody = !putHeaderInBody && Body.size() == 0 && Footer.size() > 0;
    for( size_t i = 0; i < Header.size(); i++ )
        Header[i].translate( putHeaderInBody ? mifTbl.TblBody : mifTbl.TblH, mifDoc );
    for( size_t i = 0; i < Body.size(); i++)
        Body[i].translate( mifTbl.TblBody, mifDoc );
    for( size_t i = 0; i < Footer.size(); i++ )
        Footer[i].translate( putFooterInBody ? mifTbl.TblBody : mifTbl.TblF, mifDoc );
}

void MifFOTBuilder::TablePart::processColumns() {

    MifDoc::Tbl &mifTbl = mifTable( *MifDoc::CurInstance );
    mifTbl.setTblNumColumns( 0 );
    mifTbl.setTblNumColumns( Columns.size() );
    mifTbl.TblColumnWidths.resize( 0 );
    mifTbl.TblColumnWidths.resize( Columns.size() );
    mifTbl.TblColumnWidthsAreSet();

    long totalNonproportionalWidth = 0L;
    double totalProportionalUnits = 0.0;
    for( size_t i = 0; i < Columns.size(); i++ ) {
        if( Columns[i].hasWidth ) {
            if( Columns[i].width.tableUnitFactor ) {
                totalProportionalUnits += Columns[i].width.tableUnitFactor;
            } else {
                mifTbl.TblColumnWidths[i]
                 = MifFOTBuilder::curInstance().computeLengthSpec( Columns[i].width );
                totalNonproportionalWidth += mifTbl.TblColumnWidths[i];
            }
        }
    }

    double proportionalUnit = 0.0;
    if( totalProportionalUnits )
        proportionalUnit
         = ( parentTable().tableWidth - totalNonproportionalWidth ) / totalProportionalUnits;
    
    for( size_t i = 0; i < Columns.size(); i++ ) {
        if( Columns[i].hasWidth ) {
            if( Columns[i].width.tableUnitFactor )
                mifTbl.TblColumnWidths[i]
                 = long(proportionalUnit * Columns[i].width.tableUnitFactor);
        } else 
            mifTbl.TblColumnWidths[i] = long(proportionalUnit);
    }

    columnsProcessed = true;
}

// --------- MifFOTBuilder::Row -----------------------------------------------

void MifFOTBuilder::Row::translate( Vector<MifDoc::Row> &mifRows, MifDoc &mifDoc ) {

    mifRows.resize( mifRows.size() + 1 );
    MifDoc::Row &mifRow = mifRows.back();
    mifRow.Cells.resize( Cells.size() - 1 );
    for( size_t i = 0; i + 1 < Cells.size(); i++ ) {
        Cells[i].translate( Cells[i].mifCell(), mifDoc );
        mifRow.Cells[i] = Cells[i].mifCell();    
    }
}

// --------- MifFOTBuilder::Cell ----------------------------------------------

void MifFOTBuilder::Cell::translate( MifDoc::Cell &mifCell, MifDoc &mifDoc ) {

   MifDoc::T_tagstring rulingTag; 

   rulingTag = beforeRowBorder.makeMifRuling( mifDoc );
   if( rulingTag.size() )
       mifCell.setCellTRuling( rulingTag );

   rulingTag = afterRowBorder.makeMifRuling( mifDoc );
   if( rulingTag.size() )
       mifCell.setCellBRuling( rulingTag );

   rulingTag = beforeColumnBorder.makeMifRuling( mifDoc );
   if( rulingTag.size() )
       mifCell.setCellLRuling( rulingTag );

   rulingTag = afterColumnBorder.makeMifRuling( mifDoc );
   if( rulingTag.size() )
       mifCell.setCellRRuling( rulingTag );
}

// --------- MifFOTBuilder::Border --------------------------------------------

void MifFOTBuilder::Border::setFromFot() {

    MifFOTBuilder::Format &f = MifFOTBuilder::curInstance().format();
    borderPriority = f.FotBorderPriority;
    borderPresent = f.FotBorderPresent;
    lineThickness = f.FotLineThickness;
    lineRepeat = f.FotLineRepeat;
    lineSep  = f.FotLineSep;
}

MifDoc::T_tagstring MifFOTBuilder::Border::makeMifRuling( MifDoc &mifDoc ) {

    MifDoc::T_tagstring result;

    if( borderPresent ) {

        MifDoc::Ruling *mifRuling = new MifDoc::Ruling;
    
        mifRuling->setRulingPenWidth( lineThickness );
        mifRuling->setRulingLines( lineRepeat >= 2 ? 2 : lineRepeat );
        mifRuling->setRulingGap( lineRepeat >= 2 ? lineSep - lineThickness : 0 );
    
        result = MifDoc::T_tagstring( mifRuling->key( *mifRuling ) );
        mifRuling->setRulingTag( result );

        if( mifDoc.rulingCatalog().Rulings.lookup( result ) == NULL ) {
            mifDoc.rulingCatalog().Rulings.insert( mifRuling );    
        } else
            delete mifRuling;

    }

    return result;
}

void MifFOTBuilder::Border::resolve( Border &adjacentBorder ) {
    
    if( adjacentBorder.borderPriority > borderPriority ) {

        lineThickness = adjacentBorder.lineThickness;
        borderPresent = adjacentBorder.borderPresent;
        lineRepeat = adjacentBorder.lineRepeat;
        lineSep = adjacentBorder.lineSep;

        if( adjacentBorder.cellBorder )
            adjacentBorder.borderPresent = false;
    }
}

//-----------------------------------------------------------------------------

void MifFOTBuilder::setFontSize( long n ) {

    nextFormat.FSize = n;
}

void MifFOTBuilder::setLineSpacing( const LengthSpec &l ) {

    nextFormat.FotLineSpacingSpec = l;
}

void MifFOTBuilder::setFontWeight( Symbol weight ) {

    if( weight > symbolMedium ) {
        nextFormat.FWeight = MifDoc::sBold;
    } else {
        nextFormat.FWeight = MifDoc::sRegular;
    }
}

void MifFOTBuilder::setFontPosture( Symbol posture ) {

    switch( posture ) {

      case symbolOblique:
      case symbolBackSlantedOblique:
      case symbolItalic:
      case symbolBackSlantedItalic:

        nextFormat.FAngle = MifDoc::sItalic;
        break;

      default:
        nextFormat.FAngle = MifDoc::sRegular;
    }
}

inline void MifFOTBuilder::setStartIndent( const LengthSpec &l ) {

    nextFormat.FotStartIndentSpec = l;
}

inline void MifFOTBuilder::setEndIndent( const LengthSpec &l ) {

    nextFormat.FotEndIndentSpec = l;
}

inline void MifFOTBuilder::setFirstLineStartIndent( const LengthSpec &l ) {

    nextFormat.FotFirstLineStartIndentSpec = l;
}

void MifFOTBuilder::setQuadding( Symbol quadding ) {

    switch( quadding ) {

      case symbolEnd:

        nextFormat.PgfAlignment = MifDoc::sRight;
        break;

      case symbolCenter:

        nextFormat.PgfAlignment = MifDoc::sCenter;
        break;

      case symbolJustify:

        nextFormat.PgfAlignment = MifDoc::sLeftRight;
        break;

      default: // symbolStart
        nextFormat.PgfAlignment = MifDoc::sLeft;
    }
}

inline void MifFOTBuilder::setKern( bool kern ) {

    nextFormat.FPairKern = kern;
}

inline void MifFOTBuilder::setLanguage( Letter2 code ) {

    nextFormat.FotLanguage = code;
    nextFormat.computePgfLanguage();
}

inline void MifFOTBuilder::setCountry( Letter2 code ) {

    nextFormat.FotCountry = code;
    nextFormat.computePgfLanguage();
}

inline void MifFOTBuilder::setHyphenate( bool hyphenate ) {

    nextFormat.PgfHyphenate = hyphenate;
}

inline void MifFOTBuilder::setHyphenationRemainCharCount( long n ) {

    nextFormat.HyphenMinPrefix = n;
}

inline void MifFOTBuilder::setHyphenationPushCharCount( long n ) {

    nextFormat.HyphenMinSuffix = n;
}

inline void MifFOTBuilder::setHyphenationLadderCount( long n ) {

    nextFormat.HyphenMaxLines = n;
}

void MifFOTBuilder::setFontFamilyName( const StringC &s ) {

    nextFormat.FFamily.resize( 0 );
    for( size_t i = 0; i < s.size(); i++ )
        nextFormat.FFamily += char(s[i]);
}

inline void MifFOTBuilder::setWidowCount( long n ) {

    nextFormat.PgfBlockSize = n;
}

inline void MifFOTBuilder::setOrphanCount( long n ) {

    nextFormat.PgfBlockSize = n;
}

inline void MifFOTBuilder::setPageWidth( Length pWidth ) {

    nextFormat.FotPageHeight = pWidth;
}

inline void MifFOTBuilder::setPageHeight( Length pHeight ) {

    nextFormat.FotPageHeight = pHeight;
}

inline void MifFOTBuilder::setLeftMargin( Length leftM ) {

    nextFormat.FotLeftMargin = leftM;
}

inline void MifFOTBuilder::setRightMargin( Length rightM ) {

    nextFormat.FotRightMargin = rightM;
}

inline void MifFOTBuilder::setTopMargin( Length topM ) {

    nextFormat.FotTopMargin = topM;
}

inline void MifFOTBuilder::setBottomMargin( Length bottomM ) {

    nextFormat.FotBottomMargin = bottomM;
}

void MifFOTBuilder::setHeaderMargin( Length headerM ) {

    nextFormat.FotHeaderMargin = headerM;
}

void MifFOTBuilder::setFooterMargin( Length footerM ) {

    nextFormat.FotFooterMargin = footerM;
}

void MifFOTBuilder::setDisplayAlignment( Symbol alignment ) {

    nextFormat.FotDisplayAlignment = alignment;
}

//--------- LEADERS -----------------------------------------------------------

void MifFOTBuilder::startLeader( const LeaderNIC & ) {

    checkForParagraphReopening();
    lastFlowObjectWasWhitespace = false;

    mifDoc.outSpecialChar( MifDoc::sTab );
    inLeader = true;
    setCurLeaderStream( new MifTmpOutputByteStream() );
}

void MifFOTBuilder::endLeader() {

    if( !mifDoc.curPara()->leaderTabsSet ) {
 
        String<char> leaderStr;
        curLeaderStream().commit( leaderStr );

        int numTabs = mifDoc.curFormat().PgfNumTabs + 1;
        mifDoc.curFormat().setPgfNumTabs( numTabs );
        mifDoc.curPara()->format().TabStops.push_back
         ( MifDoc::TabStop( MifDoc::sRight, format().FotCurDisplaySize
                                             - mifDoc.curFormat().PgfRIndent,
                            MifDoc::T_string( leaderStr ) ) );
        mifDoc.curPara()->format().setPgfNumTabs( numTabs );
        mifDoc.curPara()->curFormat().setPgfNumTabs( numTabs );
        mifDoc.curPara()->format().setProperties |= MifDoc::ParagraphFormat::fTabStops;

        mifDoc.curPara()->leaderTabsSet = true;
    }

    if( CurLeaderStream ) {
        delete CurLeaderStream;
        CurLeaderStream = NULL;
    }
    inLeader = false;
    lastFlowObjectWasWhitespace = false;
}

//--------- LINE FIELDS -------------------------------------------------------

void MifFOTBuilder::startLineField( const LineFieldNIC & ) {

    checkForParagraphReopening();
    lastFlowObjectWasWhitespace = false;

    Length fieldWidth = computeLengthSpec( nextFormat.FotFieldWidth );
    Length firstLineIndent = mifDoc.curPara()->format().PgfFIndent;
    bool leadingTab = true;
    switch( nextFormat.FotFieldAlign ) {
        case symbolStart: default:
            mifDoc.curPara()->format().TabStops.push_back
             ( MifDoc::TabStop( MifDoc::sLeft, firstLineIndent + fieldWidth ) );
            leadingTab = false;
            break;
        case symbolEnd:
            mifDoc.curPara()->format().TabStops.push_back
             ( MifDoc::TabStop( MifDoc::sRight, firstLineIndent + fieldWidth ) );
            mifDoc.curPara()->format().TabStops.push_back
             ( MifDoc::TabStop( MifDoc::sLeft, firstLineIndent + fieldWidth + 1 ) );
            break;
        case symbolCenter:
            mifDoc.curPara()->format().TabStops.push_back
             ( MifDoc::TabStop( MifDoc::sCenter, firstLineIndent + fieldWidth / 2 ) );
            mifDoc.curPara()->format().TabStops.push_back
             ( MifDoc::TabStop( MifDoc::sLeft, firstLineIndent + fieldWidth + 1 ) );
    }


    int numTabs = mifDoc.curFormat().PgfNumTabs + leadingTab ? 2 : 1;
    mifDoc.curFormat().setPgfNumTabs( numTabs );
    mifDoc.curPara()->format().setPgfNumTabs( numTabs );
    mifDoc.curPara()->curFormat().setPgfNumTabs( numTabs );
    mifDoc.curPara()->format().setProperties |= MifDoc::ParagraphFormat::fTabStops;

    start();

    if( leadingTab )
        mifDoc.outSpecialChar( MifDoc::sTab );
}

void MifFOTBuilder::endLineField() {

    mifDoc.outSpecialChar( MifDoc::sTab );
    end();
    lastFlowObjectWasWhitespace = false;
}

//--------- PARAGRAPH BREAKS --------------------------------------------------

void MifFOTBuilder::paragraphBreak( const ParagraphNIC &nic ) {

    if( MifDoc::Para::currentlyOpened ) {
        if( paragraphBreakInEffect )
            doEndParagraph( false, false, false );
        else {
            doEndParagraph( true, false, false );
            paragraphBreakInEffect = true;
        }
        doStartParagraph( nic );
    }
}

//--------- SCORES ------------------------------------------------------------

void MifFOTBuilder::startScore( Symbol scoreType ) {

    checkForParagraphReopening();

    switch( scoreType ) {
        case symbolBefore:
            nextFormat.setFOverline( true );
            break;
        case symbolThrough:
            nextFormat.setFStrike( true );
            break;
        case symbolAfter: default:
            nextFormat.setFUnderlining
             ( nextFormat.FotLineRepeat > 1 ? MifDoc::sFDouble : MifDoc::sFSingle );
            break;
    }

    start();
}

void MifFOTBuilder::endScore() {

    end();
}

//--------- ANCHORED FRAMES (GENERAL) -----------------------------------------

MifDoc::Frame &MifFOTBuilder::makeAnchoredFrame
 ( MifDoc::T_keyword frameType, long width, long height, MifDoc::T_keyword anchorAlign ) {

    mifDoc.aFrames().resize( mifDoc.aFrames().size() + 1 );
    MifDoc::Frame &frame = mifDoc.aFrames().back();

    frame.setFrameType( frameType );
    frame.setAnchorAlign( anchorAlign );
    frame.setShapeRect( MifDoc::T_LTWH( 0,  0, width, height ) );
 
    return frame;
}

//--------- EXTERNAL GRAPHICS -------------------------------------------------

void MifFOTBuilder::externalGraphic( const ExternalGraphicNIC &nic ) {

    bool isInline = MifDoc::Para::currentlyOpened ? true : false;

    if( !isInline )
        startDisplay( nic );
    start();

    MifDoc::T_pathname mifPathname;
    if( systemIdToMifPathname( nic.entitySystemId, mifPathname ) ) {


        MifDoc::T_keyword mifAlignment = MifDoc::sLeft;
        switch( format().FotDisplayAlignment ) {
            case symbolStart:   mifAlignment = MifDoc::sLeft; break;
            case symbolEnd:     mifAlignment = MifDoc::sRight; break;
            case symbolCenter:  mifAlignment = MifDoc::sCenter; break;
            case symbolInside:  mifAlignment = MifDoc::sInside; break;
            case symbolOutside: mifAlignment = MifDoc::sOutside; break;
            default: assert( false );
        }

        MifDoc::Frame &frame
         = makeAnchoredFrame( isInline ? MifDoc::sInline : MifDoc::sBelow,
                              nic.hasMaxWidth ? computeLengthSpec( nic.maxWidth ) : 72000,
                              nic.hasMaxHeight ? computeLengthSpec( nic.maxHeight ) : 72000,
                              mifAlignment );

        MifDoc::ImportObject *importObject
         = new MifDoc::ImportObject( mifPathname, frame.ShapeRect );        
        frame.Objects.resize( frame.Objects.size() + 1 );
        frame.Objects.back() = importObject;

        if( !isInline )
            doStartParagraph( nic, true, 0, false );
        else
            checkForParagraphReopening();
        mifDoc.outAFrame( frame.ID, mifDoc.os() );
        if( !isInline )
            endParagraph();
        else
            lastFlowObjectWasWhitespace = false;
    }

    end();
    if( !isInline )
        endDisplay();
}

enum TComponent { cName, cUp, cRoot, cRootDrive };
void addComponent( MifDoc::T_pathname &target, TComponent cType, StringC &component ) {

    target += '<';
    target += cType == cName ? 'c' : ( cType == cUp ? 'u' : 'r' );
    target += '\\';
    target += '>';
    if( cType == cName || cType == cRootDrive )
        for( size_t i = 0; i < component.size(); i++ )
            target += char( component[i] );
}

bool MifFOTBuilder::systemIdToMifPathname
 ( const StringC &systemId, MifDoc::T_pathname &mifPathname ) {

    StringC filename;
    StringC component;
    int result;

    if( ( result = systemIdFilename( systemId, filename ) ) < 0 ) {
        App->message( MifMessages::systemIdNotFilename, StringMessageArg( systemId ) );
        return false;
    } else {
        bool firstComponent = true;
        size_t i = 0;
        do {
            component.resize( 0 );
            while( i < filename.size() && filename[i] != '\\' && filename[i] != '/' )
                    component += char( filename[i++] );
            switch( component.size() ) {
                case 2:
                    if( firstComponent && component[1] == ':' )
                        addComponent( mifPathname, cRootDrive, component );
                    else
                    if( component[0] == '.' && component[1] == '.' )
                        addComponent( mifPathname, cUp, component );
                    else 
                        goto add_component;
                    break;
                case 1:
                    if( component[0] != '.' )
                        goto add_component;
                    break;
                case 0:
                    if( firstComponent && filename.size() > 0 )
                        addComponent( mifPathname, cRoot, component );
                    break;
                add_component: default:
                    addComponent( mifPathname, cName, component );
            }
            firstComponent = false;
            i++;
        } while( i < filename.size() );
    }

    return result == 0 ? false : true;
}

// System Id Resolution from James Clark

// Return -1 if an error should be generated
// FIXME in some cases should copy the entity into a file in the same
// directory as the output file.

int MifFOTBuilder::systemIdFilename(const StringC &systemId, StringC &filename)
{
  if (systemId.size() == 0)
    return -1;
  Owner<InputSource> in(EntityManager->open(systemId,
					     *SystemCharset,
					     InputSourceOrigin::make(),
					     0,
					     *App));
  if (!in)
    return 0;
  Xchar c = in->get(*App);
  StorageObjectLocation soLoc;
  ParsedSystemId parsedBuf;
  if (c == InputSource::eE && in->accessError()) {
    if (!EntityManager->parseSystemId(systemId, *SystemCharset, 0, 0, *App, parsedBuf))
      return 0;
    if (parsedBuf.size() != 1 || parsedBuf[0].baseId.size())
      return 0;
    soLoc.storageObjectSpec = &parsedBuf[0];
    soLoc.actualStorageId = parsedBuf[0].specId;
  }
  else {
    const Location &loc = in->currentLocation();
    if (loc.origin().isNull())
      return -1;
    const InputSourceOrigin *tem = loc.origin()->asInputSourceOrigin();
    if (!tem)
      return -1;
    const ParsedSystemId *psi
      = ExtendEntityManager::externalInfoParsedSystemId(tem->externalInfo());
    if (!psi || psi->size() != 1)
      return -1;
    if (!ExtendEntityManager::externalize(tem->externalInfo(), 0, soLoc))
      return -1;
  }
  if (strcmp(soLoc.storageObjectSpec->storageManager->type(), "OSFILE") != 0)
    return -1;
  filename = soLoc.actualStorageId;
  return 1;
}

//--------- RULES -------------------------------------------------------------

void MifFOTBuilder::rule( const RuleNIC &nic ) {

    bool isInline = ( nic.orientation == symbolHorizontal || nic.orientation == symbolVertical )
                     ? false : true;

    if( isInline )
        checkForParagraphReopening();

    if( !isInline )
        startDisplay( nic );
    start();

    long indentlessDisplaySize
          = format().FotCurDisplaySize - computeLengthSpec( format().FotStartIndentSpec )
            - computeLengthSpec( format().FotEndIndentSpec );

    MifDoc::T_dimension ruleHeight
     = format().FotLineThickness + format().FotLineSep * ( format().FotLineRepeat - 1 );
    if( ruleHeight < 0 )
        ruleHeight = 0;
        
    MifDoc::T_dimension ruleLength;
    MifDoc::T_dimension ruleOffset;

    if( nic.hasLength && ( ruleLength = computeLengthSpec( nic.length ) ) > 0 ) {
        switch( format().FotDisplayAlignment ) {
            case symbolStart: default:
                ruleOffset = 0;
                break;
            case symbolCenter:
                ruleOffset = ( indentlessDisplaySize - ruleLength ) / 2;
                break;
            case symbolEnd:
                ruleOffset = indentlessDisplaySize - ruleLength;
        }
    } else {
        ruleOffset = 0;
        ruleLength = indentlessDisplaySize;
    }

    MifDoc::Frame &frame
     = makeAnchoredFrame( MifDoc::sInline, isInline ? ruleLength : indentlessDisplaySize,
                          ruleHeight );

   if( isInline )
        frame.setBLOffset( computeLengthSpec( format().FotPositionPointShiftSpec ) );
    else
        frame.setBLOffset( -ruleHeight / 2 + 4000 / 3 ); // min font baseline correction

    MifDoc::T_keyword capType;
    switch( format().FotLineCap ) {
        case symbolButt: default: capType = MifDoc::sButt; break;
        case symbolRound: capType = MifDoc::sRound; break;
        case symbolSquare: capType = MifDoc::sSquare;
    }
    MifDoc::PolyLine *polyLine;
    Length curLineVOffset = format().FotLineThickness / 2;
    for( long i = format().FotLineRepeat; i > 0; i--, curLineVOffset += format().FotLineSep ) {

        polyLine = new MifDoc::PolyLine( capType, 0, 0, format().FotLineThickness );        
        frame.Objects.resize( frame.Objects.size() + 1 );
        frame.Objects.back() = polyLine;

        polyLine->setHeadCap( capType );
        polyLine->setTailCap( capType );
        polyLine->Points.resize( polyLine->Points.size() + 1 );
        polyLine->Points.back() = MifDoc::T_XY( ruleOffset, curLineVOffset );
        polyLine->Points.resize( polyLine->Points.size() + 1 );
        polyLine->Points.back() = MifDoc::T_XY( ruleOffset + ruleLength, curLineVOffset );
    }

    if( !isInline ) doStartParagraph( nic, true, 0, true );
    mifDoc.outAFrame( frame.ID, mifDoc.os() );
    if( !isInline )
        endParagraph();
    else
        lastFlowObjectWasWhitespace = false;

    end();
    if( !isInline )
        endDisplay();
}

//--------- TABLES ------------------------------------------------------------

void MifFOTBuilder::startTable( const TableNIC &nic ) {

    startDisplay( nic );
    start();

    if( !curTable().DefaultTblFormatGenerated ) {
    
        MifDoc::TblFormat defaultTblFormat( MifDoc::sDefaultTblFormat );
        defaultTblFormat.setDSSSLDefaults();
        mifDoc.tblCatalog().TblFormats.push_back( defaultTblFormat );
        curTable().DefaultTblFormatGenerated = true;
    }

    Length curStartIndent = computeLengthSpec( format().FotStartIndentSpec );
    curTable().startIndent = curStartIndent;

    curTable().begin( mifDoc );
//    curTable().nic = nic;
    curTable().displayAlignment = format().FotDisplayAlignment;

    if( nic.widthType == TableNIC::widthExplicit )
        curTable().tableWidth = computeLengthSpec( nic.width );
    else
        curTable().tableWidth
         = format().FotCurDisplaySize - curStartIndent
            - computeLengthSpec( format().FotEndIndentSpec );


    doStartParagraph( nic, true, 0 );
    endParagraph();
}

void MifFOTBuilder::endTable() {

    bool firstPart, lastPart, hasHeader, hasFooter;
    for( size_t i = 0; i < curTable().TableParts.size(); i++ ) {
        firstPart = ( i == 0 ) ? true : false;
        lastPart = ( i == curTable().TableParts.size() - 1 ) ? true : false;
        TablePart &tablePart = curTable().TableParts[i];
        hasHeader = tablePart.Header.size() > 0 ? true : false;
        hasFooter = tablePart.Footer.size() > 0 ? true : false;
        if( hasHeader )
            curTable().resolveBorders( tablePart.Header,
                                       firstPart ? true : false, false );
        curTable().resolveBorders( tablePart.Body,
                                   hasHeader ? false : true, hasFooter ? false : true );
        if( hasFooter )
            curTable().resolveBorders( tablePart.Footer,
                                       false, lastPart ? true : false );
        tablePart.translate( mifDoc );
    }

    MifDoc::CurInstance->setCurTblNum( 0 );

    endDisplay();
    end();
}

void MifFOTBuilder::tableColumn( const TableColumnNIC &nic ) {

    if( nic.columnIndex >= curTable().curTablePart().Columns.size() )
        curTable().curTablePart().Columns.resize( nic.columnIndex + 1 );
  
    curTable().curTablePart().Columns[nic.columnIndex].hasWidth = nic.hasWidth;
    if( nic.hasWidth ) {
        curTable().curTablePart().Columns[nic.columnIndex].width = nic.width;
    }
}

void MifFOTBuilder::startTableRow() {

    curTable().curRows().resize( curTable().curRows().size() + 1 );
}

void MifFOTBuilder::endTableRow() {

}

void MifFOTBuilder::startTableCell( const TableCellNIC &nic ) {

    start();

    TablePart &tp = curTable().curTablePart();
    if( !tp.columnsProcessed )
        tp.processColumns();
   
    Vector<Cell> &Cells = curTable().curRows().back().Cells;
    if( nic.columnIndex >= Cells.size() )
        Cells.resize( nic.columnIndex + 1 );

    Cell &cell = Cells[nic.columnIndex];
    curTable().CurCell = &cell;
    cell.missing = nic.missing;

    if( nic.nColumnsSpanned != 1 ) {
        cell.nColumnsSpanned = nic.nColumnsSpanned;
        cell.mifCell().setCellColumns( nic.nColumnsSpanned );
    }

    if( nic.nRowsSpanned != 1 ) {
        cell.nRowsSpanned = nic.nRowsSpanned;
        cell.mifCell().setCellRows( nic.nRowsSpanned );
    }

    long newDisplaySize = 0;
    for( size_t i = nic.columnIndex; i < nic.columnIndex + nic.nColumnsSpanned; i++ )
        if( i < tp.Columns.size() ) {
            if( tp.Columns[i].hasWidth )
                newDisplaySize
                 += computeLengthSpec( tp.Columns[i].width );
        } else
        if( !nic.missing ) {
            App->message( MifMessages::missingTableColumnFlowObject );  
            // NOTE: at this point there's already a danger of not realizing
            // right display space sizes inside cells
            tp.Columns.resize( tp.Columns.size() + 1 ); 
            tp.Columns.back().hasWidth = true;
            TableLengthSpec tls; tls.tableUnitFactor = 1.0;
            tp.Columns.back().width = tls;
            tp.needsColumnReprocessing = true;
            if( i > tp.mifTable( mifDoc ).TblNumColumns )
                tp.mifTable( mifDoc ).setTblNumColumns( i );
        }

    newDisplaySize -= format().PgfCellMargins.l + format().PgfCellMargins.r;

    if( newDisplaySize > 0 )
        format().FotCurDisplaySize = newDisplaySize;

    mifDoc.enterTableCell( cell.mifCell() ); 
}

void MifFOTBuilder::endTableCell() {

    mifDoc.exitTableCell();
    end();

    curTable().CurCell = NULL;
}

void MifFOTBuilder::startTablePartSerial( const TablePartNIC &nic ) {

    startDisplay( nic );
    start();

    if( curTable().NoTablePartsSeen )
        curTable().NoTablePartsSeen = false;
    else
        curTable().TableParts.resize( curTable().TableParts.size() + 1 );

    curTable().TableParts.back().begin( curTable(), mifDoc );    

    doStartParagraph( nic, true, 0 );
    endParagraph();
}
 
void MifFOTBuilder::endTablePartSerial() {

    curTable().CurTablePart = NULL;
    endDisplay();
    end();
}

void MifFOTBuilder::startTablePartHeader() {

    curTable().CurRows = &curTable().curTablePart().Header; 
}

void MifFOTBuilder::endTablePartHeader() {

    curTable().CurRows = &curTable().curTablePart().Body; 
}

void MifFOTBuilder::startTablePartFooter() {

    curTable().CurRows = &curTable().curTablePart().Footer; 
}

void MifFOTBuilder::endTablePartFooter() {

    curTable().CurRows = &curTable().curTablePart().Body; 
}

void MifFOTBuilder::setCellBeforeRowMargin( Length l ) {

    nextFormat.PgfCellMargins.t = l;
}

void MifFOTBuilder::setCellAfterRowMargin( Length l ) {

    nextFormat.PgfCellMargins.b = l;
}

void MifFOTBuilder::setCellBeforeColumnMargin( Length l ) {

    nextFormat.PgfCellMargins.l = l;
}

void MifFOTBuilder::setCellAfterColumnMargin( Length l ) {

    nextFormat.PgfCellMargins.r = l;
}

void MifFOTBuilder::setCellRowAlignment( Symbol alignment ) {

    switch( alignment ) {
        case symbolEnd:     nextFormat.PgfCellAlignment = MifDoc::sBottom; break;
        case symbolCenter:  nextFormat.PgfCellAlignment = MifDoc::sMiddle; break;
        default:            nextFormat.PgfCellAlignment = MifDoc::sTop; break;
    }
}

//--------- TABLE BORDERS -----------------------------------------------------

void MifFOTBuilder::setBorderPresent( bool b ) {

    nextFormat.FotBorderPresent = b;
}

void MifFOTBuilder::setBorderPriority( long p ) {

    nextFormat.FotBorderPriority = p;
}

void MifFOTBuilder::setBorderOmitAtBreak(bool) {
}

void MifFOTBuilder::tableBeforeRowBorder() {

    start();
    curTable().beforeRowBorder.setFromFot();
    end();
}

void MifFOTBuilder::tableAfterRowBorder() {

    start();
    curTable().afterRowBorder.setFromFot();
    end();
}

void MifFOTBuilder::tableBeforeColumnBorder() {

    start();
    curTable().beforeColumnBorder.setFromFot();
    end();
}

void MifFOTBuilder::tableAfterColumnBorder() {

    start();
    curTable().afterColumnBorder.setFromFot();
    end();
}

void MifFOTBuilder::tableCellBeforeRowBorder() {

    start();
    curTable().curCell().beforeRowBorder.setFromFot();
    end();
}

void MifFOTBuilder::tableCellAfterRowBorder() {

    start();
    curTable().curCell().afterRowBorder.setFromFot();
    end();
}

void MifFOTBuilder::tableCellBeforeColumnBorder() {

    start();
    curTable().curCell().beforeColumnBorder.setFromFot();
    end();
}

void MifFOTBuilder::tableCellAfterColumnBorder() {

    start();
    curTable().curCell().afterColumnBorder.setFromFot();
    end();
}

//--------- CROSS-REFERENCES --------------------------------------------------

void MifFOTBuilder::outPotentialMarkers() {

  for( size_t i = 0; i < nodeStack.size(); i++ ) {
      unsigned long n;
      if( nodeStack[i].node->elementIndex( n ) == accessOK ) {
          unsigned long groveIndex = nodeStack[i].node->groveIndex();
          mifDoc.os() << MifDoc::PotentialMarkerInfo
                          ( groveIndex, n, mifDoc.os().CurTagIndent );
          mifDoc.elements().setBookComponentIndex
           ( groveIndex, n, mifDoc.bookComponents().size() - 1 );
      }
  }
  NodeInfo::nonEmptyElementsOpened = 0;
  nodeStack.resize(0);
}

void MifFOTBuilder::startNode( const NodePtr &node, const StringC &mode, RuleType ) {

    NodeInfo::curNodeLevel++;
    if( mode.size() == 0 )
        nodeStack.push_back( NodeInfo( node, NodeInfo::curNodeLevel ) );
}

void MifFOTBuilder::endNode() {

    if( nodeStack.size() > 0 && nodeStack.back().nodeLevel == NodeInfo::curNodeLevel
         && NodeInfo::nonEmptyElementsOpened < nodeStack.size() )
        nodeStack.resize( nodeStack.size() - 1 );
    NodeInfo::curNodeLevel--;
}

void MifFOTBuilder::currentNodePageNumber( const NodePtr &node ) {

    unsigned long n;
    if( node->elementIndex( n ) == accessOK ) {
        if( !mifDoc.bookComponent().pageNumXRefFormatGenerated ) {
            mifDoc.bookComponent().XRefFormats.push_back
             ( MifDoc::XRefFormat( MifDoc::sPageNumXRefFormatName, "<$pagenum\\>" ) );
            mifDoc.bookComponent().pageNumXRefFormatGenerated = true;
        }
        unsigned long groveIndex = node->groveIndex();
        mifDoc.os() << MifDoc::XRefInfo( groveIndex, n, mifDoc.os().CurTagIndent );
        mifDoc.elements().setReferencedFlag( groveIndex, n );
    }
}

//--------- INDEX-ENTRIES -----------------------------------------------------

void MifFOTBuilder::indexEntry( const IndexEntryNIC &nic ) {

    MifDoc::T_string mText;
    if( nic.components.size() > 0 ) {
        if( !nic.pageNumber )
            mText += MifDoc::T_string( "<$nopage\\>" );
        if( nic.startsPageRange )
            mText += MifDoc::T_string( "<$startrange\\>" );
        if( nic.endsPageRange )
            mText += MifDoc::T_string( "<$endrange\\>" );
        bool first = true;
        for( size_t i = 0; i < nic.components.size(); first = false, i++ ) {
            if( !first ) mText += ':';
            outString( nic.components[i].data(), nic.components[i].size(), NULL, false, &mText ); 
        }
        if( nic.sortString.size() > 0 ) {
            mText += '[';
            outString( nic.sortString.data(), nic.sortString.size(), NULL, false, &mText ); 
            mText += ']';
        }
        MifDoc::Marker marker( mText );
        marker.setMType( 2 );
        marker.out( mifDoc.os() );
    }
}

//-----------------------------------------------------------------------------

void MifFOTBuilder::setSpan( long n ) {

    nextFormat.FotSpan = n;
}

void MifFOTBuilder::setPageNColumns( long n ) {

    nextFormat.FotPageNColumns = n;
    mifDoc.document().setDColumns( n );
}

void MifFOTBuilder::setPageColumnSep( Length l ) {

    nextFormat.FotPageColumnSep = l;
    mifDoc.document().setDColumnGap( l );
}

void MifFOTBuilder::setPageBalanceColumns( bool b ) {

    nextFormat.FotPageBalanceColumns = b;
}

void MifFOTBuilder::setLines( Symbol s ) {

    nextFormat.FotLines = s;
}

void MifFOTBuilder::setInputWhitespaceTreatment( Symbol s ) {

    nextFormat.FotInputWhitespaceTreatment = s;
}

void MifFOTBuilder::setMinLeading( const OptLengthSpec &ol ) {

    nextFormat.FotMinLeading = ol;
}

void MifFOTBuilder::setPositionPointShift( const LengthSpec &l ) {

    nextFormat.FotPositionPointShiftSpec = l;
}

void MifFOTBuilder::setLineThickness( Length l ) {

    nextFormat.FotLineThickness = l;
}

void MifFOTBuilder::setLineRepeat( long n ) {

    nextFormat.FotLineRepeat = n;
}

void MifFOTBuilder::setLineSep( Length l ) {

    nextFormat.FotLineSep = l;
}

void MifFOTBuilder::setFieldWidth( const LengthSpec &l ) {

    nextFormat.FotFieldWidth = l;
}

void MifFOTBuilder::setLineCap( Symbol s ) {

    nextFormat.FotLineCap = s;
}

void MifFOTBuilder::setFieldAlign( Symbol s ) {

    nextFormat.FotFieldAlign = s;
}

//---------------------------------------------------------------------------------------------

void MifFOTBuilder::extension( const ExtensionFlowObj &fo, const NodePtr &node ) {

  ((const MifExtensionFlowObj &)fo).atomic( *this, node );
}

bool MifFOTBuilder::IndexEntryFlowObj::hasNIC( const StringC &name ) const {

    return name == "components" || name == "page-number?" || name == "sort-string"
           || name == "starts-page-range?" || name == "ends-page-range?" ;
}

void MifFOTBuilder::IndexEntryFlowObj::setNIC( const StringC &name, const Value &value ) {

    if( name == "components" )
        value.convertStringList( nic.components );
    else if( name == "page-number?" )
        value.convertBoolean( nic.pageNumber );
    else if( name == "sort-string" ) {
        bool hasString;
        value.convertBoolean( hasString );
        if( hasString )
            value.convertString( nic.sortString );
    } else if( name == "starts-page-range?" )
        value.convertBoolean( nic.startsPageRange );
    else if( name == "ends-page-range?" )
        value.convertBoolean( nic.endsPageRange );
}

//---------------------------------------------------------------------------------------------

unsigned long StringHash::hash( const String<char> &str ) {

  const char *p = str.data();
  unsigned long h = 0;
  for( size_t n = str.size(); n > 0; n-- )
    h = (h << 5) + h + *p++;	// from Chris Torek
  return h;
}

FOTBuilder *makeMifFOTBuilder( const String<CmdLineApp::AppChar> &fileLoc,
                               const Ptr<ExtendEntityManager> &entityManager,
 			                   const CharsetInfo &systemCharset,
                               CmdLineApp *app,
                               const FOTBuilder::Extension *&ext ) {

    
    MifFOTBuilder::IndexEntryFlowObj *indexEntryFlowObject
     = new MifFOTBuilder::IndexEntryFlowObj;

    static const FOTBuilder::Extension extensions[] = {
    {
      "UNREGISTERED::ISOGEN//Flow Object Class::index-entry",
      0,
      0,
      0,
      0,
      indexEntryFlowObject
    },
    {
      "UNREGISTERED::James Clark//Characteristic::page-n-columns",
      0,
      0,
      (void (FOTBuilder::*)(long))&MifFOTBuilder::setPageNColumns,
      0
    },
    {
      "UNREGISTERED::James Clark//Characteristic::page-column-sep",
      0,
      0,
      0,
      (void (FOTBuilder::*)(FOTBuilder::Length))&MifFOTBuilder::setPageColumnSep,
    },
    {
      "UNREGISTERED::James Clark//Characteristic::page-balance-columns?",
      (void (FOTBuilder::*)(bool))&MifFOTBuilder::setPageBalanceColumns,
      0,
      0
    },
    { 0, 0, 0}
  };
  ext = extensions;
  return new MifFOTBuilder( fileLoc, entityManager, systemCharset, app );
}

const MifFOTBuilder::SymbolFont
MifFOTBuilder::SymbolFonts[MifFOTBuilder::nSymbolFonts] = {
  { "Symbol", {
    0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000,
    0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000,
    0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000,
    0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000,
    0x0020, 0x0021, 0x2200, 0x0023, 0x2203, 0x0025, 0x0026, 0x220B,
    0x0028, 0x0029, 0x2217, 0x002B, 0x002C, 0x2212, 0x002E, 0x002F,
    0x0030, 0x0031, 0x0032, 0x0033, 0x0034, 0x0035, 0x0036, 0x0037,
    0x0038, 0x0039, 0x003A, 0x003B, 0x003C, 0x003D, 0x003E, 0x003F,
    0x2245, 0x0391, 0x0392, 0x03A7, 0x2206, 0x0395, 0x03A6, 0x0393,
    0x0397, 0x0399, 0x03D1, 0x039A, 0x039B, 0x039C, 0x039D, 0x039F,
    0x03A0, 0x0398, 0x03A1, 0x03A3, 0x03A4, 0x03A5, 0x03C2, 0x2126,
    0x039E, 0x03A8, 0x0396, 0x005B, 0x2234, 0x005D, 0x22A5, 0x005F,
    0x203E, 0x03B1, 0x03B2, 0x03C7, 0x03B4, 0x03B5, 0x03C6, 0x03B3,
    0x03B7, 0x03B9, 0x03D5, 0x03BA, 0x03BB, 0x03BC, 0x03BD, 0x03BF,
    0x03C0, 0x03B8, 0x03C1, 0x03C3, 0x03C4, 0x03C5, 0x03D6, 0x03C9,
    0x03BE, 0x03C8, 0x03B6, 0x007B, 0x007C, 0x007D, 0x223C, 0x0000,
    0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000,
    0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000,
    0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000,
    0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000,
    0x0000, 0x03D2, 0x2032, 0x2264, 0x2215, 0x221E, 0x0192, 0x2663,
    0x2666, 0x2665, 0x2660, 0x2194, 0x2190, 0x2191, 0x2192, 0x2193,
    0x00B0, 0x00B1, 0x2033, 0x2265, 0x00D7, 0x221D, 0x2202, 0x2022,
    0x00F7, 0x2260, 0x2261, 0x2248, 0x2026, 0x0000, 0x0000, 0x21B5,
    0x2135, 0x2111, 0x211C, 0x2118, 0x2297, 0x2295, 0x2205, 0x2229,
    0x222A, 0x2283, 0x2287, 0x2284, 0x2282, 0x2286, 0x2208, 0x2209,
    0x2220, 0x2207, 0x00AE, 0x00A9, 0x2122, 0x220F, 0x221A, 0x22C5,
    0x00AC, 0x2227, 0x2228, 0x21D4, 0x21D0, 0x21D1, 0x21D2, 0x21D3,
    0x25CA, 0x2329, 0x00AE, 0x00A9, 0x2122, 0x2211, 0x0000, 0x0000,
    0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000,
    0x0000, 0x232A, 0x222B, 0x2320, 0x0000, 0x2321, 0x0000, 0x0000,
    0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000 }
  },
  // I created this table by hand.  It's sometimes not obvious what
  // the appropriate mapping is (James Clark)
  { "Wingdings", {
    0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000,
    0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000,
    0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000,
    0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000,
    0x0000, 0x0000, 0x2702, 0x2701, 0x0000, 0x0000, 0x0000, 0x0000,
    0x260e, 0x2706, 0x2709, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000,
    0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x2328,
    0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x2707, 0x270d,
    0x0000, 0x270c, 0x0000, 0x0000, 0x0000, 0x261c, 0x261e, 0x261d,
    0x261f, 0x0000, 0x263a, 0x0000, 0x2639, 0x0000, 0x2620, 0x0000,
    0x0000, 0x2708, 0x263c, 0x0000, 0x2744, 0x0000, 0x271e, 0x0000,
    0x2720, 0x2721, 0x262a, 0x262f, 0x0950, 0x2638, 0x2648, 0x2649,
    0x264a, 0x264b, 0x264c, 0x264d, 0x264e, 0x264f, 0x2650, 0x2651,
    0x2652, 0x2653, 0x0000, 0x0000, 0x25cf, 0x274d, 0x25a0, 0x25a1,
    0x0000, 0x2751, 0x2752, 0x0000, 0x0000, 0x25c6, 0x2756, 0x0000,
    0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000,
    0x0000, 0x2780, 0x2781, 0x2782, 0x2783, 0x2784, 0x2785, 0x2786,
    0x2787, 0x2788, 0x2789, 0x0000, 0x278a, 0x278b, 0x278c, 0x278d,
    0x278e, 0x278f, 0x2790, 0x2791, 0x2792, 0x2793, 0x0000, 0x0000,
    0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000,
    0x0000, 0x25cb, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x25aa,
    0x0000, 0x0000, 0x2726, 0x2605, 0x2736, 0x0000, 0x2739, 0x0000,
    0x0000, 0x0000, 0x2727, 0x0000, 0x0000, 0x272a, 0x2730, 0x0000,
    0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000,
    0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000,
    0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000,
    0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x232b, 0x2326, 0x0000,
    0x27a2, 0x0000, 0x0000, 0x0000, 0x27b2, 0x0000, 0x0000, 0x0000,
    0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000,
    0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000,
    0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000,
    0x0000, 0x0000, 0x25ab, 0x2718, 0x2714, 0x2612, 0x2611, 0x0000 }
  },
  { "ZapfDingbats", {
    0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000,
    0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000,
    0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000,
    0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000,
    0x0000, 0x2701, 0x2702, 0x2703, 0x2704, 0x260E, 0x2706, 0x2707,
    0x2708, 0x2709, 0x261B, 0x261E, 0x270C, 0x270D, 0x270E, 0x270F,
    0x2710, 0x2711, 0x2712, 0x2713, 0x2714, 0x2715, 0x2716, 0x2717,
    0x2718, 0x2719, 0x271A, 0x271B, 0x271C, 0x271D, 0x271E, 0x271F,
    0x2720, 0x2721, 0x2722, 0x2723, 0x2724, 0x2725, 0x2726, 0x2727,
    0x2605, 0x2729, 0x272A, 0x272B, 0x272C, 0x272D, 0x272E, 0x272F,
    0x2730, 0x2731, 0x2732, 0x2733, 0x2734, 0x2735, 0x2736, 0x2737,
    0x2738, 0x2739, 0x273A, 0x273B, 0x273C, 0x273D, 0x273E, 0x273F,
    0x2740, 0x2741, 0x2742, 0x2743, 0x2744, 0x2745, 0x2746, 0x2747,
    0x2748, 0x2749, 0x274A, 0x274B, 0x0000, 0x274D, 0x25A0, 0x274F,
    0x2750, 0x2751, 0x2752, 0x25B2, 0x25BC, 0x25C6, 0x2756, 0x0000,
    0x2758, 0x2759, 0x275A, 0x275B, 0x275C, 0x275D, 0x275E, 0x0000,
    0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000,
    0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000,
    0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000,
    0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000,
    0x0000, 0x2761, 0x2762, 0x2763, 0x2764, 0x2765, 0x2766, 0x2767,
    0x2663, 0x2666, 0x2665, 0x2660, 0x2460, 0x2461, 0x2462, 0x2463,
    0x2464, 0x2465, 0x2466, 0x2467, 0x2468, 0x2469, 0x2776, 0x2777,
    0x2778, 0x2779, 0x277A, 0x277B, 0x277C, 0x277D, 0x277E, 0x277F,
    0x2780, 0x2781, 0x2782, 0x2783, 0x2784, 0x2785, 0x2786, 0x2787,
    0x2788, 0x2789, 0x278A, 0x278B, 0x278C, 0x278D, 0x278E, 0x278F,
    0x2790, 0x2791, 0x2792, 0x2793, 0x2794, 0x2192, 0x2194, 0x2195,
    0x2798, 0x2799, 0x279A, 0x279B, 0x279C, 0x279D, 0x279E, 0x279F,
    0x27A0, 0x27A1, 0x27A2, 0x27A3, 0x27A4, 0x27A5, 0x27A6, 0x27A7,
    0x27A8, 0x27A9, 0x27AA, 0x27AB, 0x27AC, 0x27AD, 0x27AE, 0x27AF,
    0x0000, 0x27B1, 0x27B2, 0x27B3, 0x27B4, 0x27B5, 0x27B6, 0x27B7,
    0x27B8, 0x27B9, 0x27BA, 0x27BB, 0x27BC, 0x27BD, 0x27BE, 0x0000 }
    },
};

#ifdef DSSSL_NAMESPACE
}
#endif

#include "MifFOTBuilder_inst.cxx"
